# Copyright 2022 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=protected-access,attribute-defined-outside-init

"""
Unit tests for _session module of the zhmcclient_mock package.
"""


from datetime import datetime

from zhmcclient import Client
from zhmcclient_mock import FakedSession

from zhmcclient_mock._hmc import \
    FakedMetricGroupDefinition, FakedMetricObjectValues

from tests.common.utils import assert_equal_hmc


# The resource input for the test_hmc_dump_load() test function.
# Each type of resource should appear in this resource input.
HMC1_RESOURCES = {
    'consoles': [
        {
            'properties': {
                'name': 'hmc1',
                'version': '2.14.1',
            },
            'users': [
                {
                    'properties': {
                        'object-id': 'user1',
                        'name': 'User 1',
                        'type': 'standard',
                    },
                },
            ],
            'user_roles': [
                {
                    'properties': {
                        'object-id': 'userrole1',
                        'name': 'User role 1',
                    },
                },
            ],
            'user_patterns': [
                {
                    'properties': {
                        'element-id': 'userpattern1',
                        'name': 'User pattern 1',
                        'type': 'glob-like',
                    },
                },
            ],
            'password_rules': [
                {
                    'properties': {
                        'element-id': 'passwordrule1',
                        'name': 'Password rule 1',
                        'type': 'user-defined',
                    },
                },
            ],
            'tasks': [
                {
                    'properties': {
                        'element-id': 'task1',
                        'name': 'Task 1',
                    },
                },
            ],
            'ldap_server_definitions': [
                {
                    'properties': {
                        'element-id': 'lsd1',
                        'name': 'LDAP server definition 1',
                    },
                },
            ],
            'unmanaged_cpcs': [
                {
                    'properties': {
                        'object-id': 'cpc3',
                        'name': 'Unmanaged CPC 3',
                    },
                },
            ],
            'storage_groups': [
                {
                    'properties': {
                        'object-id': 'sg1',
                        'name': 'Storage group 1',
                        'type': 'shared',
                        'cpc-uri': '/api/cpcs/cpc1',
                        'fulfillment-state': 'complete',
                    },
                    'storage_volumes': [
                        {
                            'properties': {
                                'element-id': 'sv1',
                                'name': 'Storage volume 1',
                                'fulfillment-state': 'complete',
                                'size': '10',
                                'usage': 'boot',
                            },
                        },
                    ],
                },
            ],
        },
    ],
    'cpcs': [
        {
            'properties': {
                'object-id': 'cpc1',
                'name': 'CPC 1 (DPM mode)',
                'dpm-enabled': True,
                'se-version': '2.13',
                'status': 'operating',
            },
            'capacity_groups': [
                {
                    'properties': {
                        'object-id': 'cg1',
                        'name': 'Capacity group 1',
                    },
                },
            ],
            'partitions': [
                {
                    'properties': {
                        'object-id': 'part1',
                        'name': 'Partition 1',
                    },
                    'nics': [
                        {
                            'properties': {
                                'element-id': 'nic1',
                                'name': 'NIC 1',
                                'device-number': '0010',
                            },
                        },
                    ],
                    'hbas': [
                        {
                            'properties': {
                                'element-id': 'hba1',
                                'name': 'HBA 1',
                                'device-number': '0018',
                            },
                        },
                    ],
                    'virtual_functions': [
                        {
                            'properties': {
                                'element-id': 'vf1',
                                'name': 'Virtual function 1',
                                'device-number': '0020',
                            },
                        },
                    ],
                },
            ],
            'adapters': [
                {
                    'properties': {
                        'object-id': 'ad1',
                        'name': 'Adapter 1',
                        'adapter-family': 'hipersockets',
                    },
                    'ports': [
                        {
                            'properties': {
                                'element-id': '1',
                                'name': 'Port 1',
                            },
                        },
                    ],
                },
            ],
            'virtual_switches': [
                {
                    'properties': {
                        'object-id': 'vs1',
                        'name': 'Virtual switch 1',
                    },
                },
            ],
        },
        {
            'properties': {
                'object-id': 'cpc2',
                'name': 'CPC 2 (classic mode)',
                'dpm-enabled': False,
                'se-version': '2.13',
                'status': 'active',
            },
            'capacity_groups': [
                {
                    'properties': {
                        'object-id': 'cg1',
                        'name': 'Capacity group 1',
                    },
                },
            ],
            'lpars': [
                {
                    'properties': {
                        'object-id': 'lpar1',
                        'name': 'LPAR1',
                    },
                },
            ],
            'reset_activation_profiles': [
                {
                    'properties': {
                        'object-id': 'rap1',
                        'name': 'RAP1',
                    },
                },
            ],
            'image_activation_profiles': [
                {
                    'properties': {
                        'object-id': 'lpar1',
                        'name': 'LPAR1',
                    },
                },
            ],
            'load_activation_profiles': [
                {
                    'properties': {
                        'object-id': 'specialload',
                        'name': 'SPECIALLOAD',
                    },
                },
            ],
        },
    ]
}

# The metroc group definitions for the test_hmc_dump_load() test function.
HMC_DUMP_LOAD_METRIC_GROUP_DEFS = [
    FakedMetricGroupDefinition(
        name='partition-usage',
        types=[
            ('processor-usage', 'integer-metric'),
            ('network-usage', 'integer-metric'),
            ('storage-usage', 'integer-metric'),
            ('accelerator-usage', 'integer-metric'),
            ('crypto-usage', 'integer-metric'),
        ]
    ),
]


def test_session_to_from_yaml():
    """Test FakedSession.to_yaml() and from_yaml()"""

    debug_yamlfile = False  # Enable to get the HMC dump file created

    # Set up the faked session
    console_props = HMC1_RESOURCES['consoles'][0]['properties']
    host = 'testhmc'
    api_version = '2.20'
    hmc_name = console_props['name']
    hmc_version = console_props['version']

    session = FakedSession(host, hmc_name, hmc_version, api_version,
                           userid='fake-user', password='fake-password')
    client = Client(session)

    session.hmc.add_resources(HMC1_RESOURCES)

    cpc1 = session.hmc.cpcs.list(filter_args={'object-id': 'cpc1'})[0]
    part1 = cpc1.partitions.list(filter_args={'object-id': 'part1'})[0]
    session.hmc.add_metric_values(
        FakedMetricObjectValues(
            group_name='partition-usage',
            resource_uri=part1.uri,
            timestamp=datetime.now(),
            values=[
                ('processor-usage', 15),
                ('network-usage', 0),
                ('storage-usage', 1),
                ('accelerator-usage', 0),
                ('crypto-usage', 0),
            ]))
    session.hmc.add_metric_values(
        FakedMetricObjectValues(
            group_name='partition-usage',
            resource_uri=part1.uri,
            timestamp=datetime.now(),
            values=[
                ('processor-usage', 17),
                ('network-usage', 5),
                ('storage-usage', 2),
                ('accelerator-usage', 0),
                ('crypto-usage', 0),
            ]))
    session.logon()  # Sets session.host

    # The function to be tested:
    hmc_yaml = client.to_hmc_yaml()

    if debug_yamlfile:
        filename = 'tmp_hmcdef.yaml'
        print(f"Creating file with faked HMC definition: {filename}")
        # pylint: disable=unspecified-encoding
        with open(filename, 'w') as fp:
            fp.write(hmc_yaml)

    # The function to be tested:
    new_session = FakedSession.from_hmc_yaml(hmc_yaml)

    assert_equal_hmc(new_session.hmc, session.hmc)



# Copyright 2017,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A :term:`User` resource represents a user configured in the HMC.
"""


import copy

from ._manager import BaseManager
from ._resource import BaseResource
from ._logging import logged_api_call
from ._utils import RC_USER

__all__ = ['UserManager', 'User']


class UserManager(BaseManager):
    """
    Manager providing access to the :term:`User` resources of a HMC.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible via the following instance variable of a
    :class:`~zhmcclient.Console` object:

    * :attr:`zhmcclient.Console.users`

    HMC/SE version requirements:

    * HMC version >= 2.13.0
    """

    def __init__(self, console):
        # This function should not go into the docs.
        # Parameters:
        #   console (:class:`~zhmcclient.Console`):
        #      Console object representing the HMC.

        # Resource properties that are supported as filter query parameters.
        # If the support for a resource property changes within the set of HMC
        # versions that support this type of resource, this list must be set up
        # for the version of the HMC this session is connected to.
        # Because this resource has case-insensitive names, this list must
        # contain the name property.
        query_props = [
            'name',
            'type',
        ]

        super().__init__(
            resource_class=User,
            class_name=RC_USER,
            session=console.manager.session,
            parent=console,
            base_uri='/api/users',
            oid_prop='object-id',
            uri_prop='object-uri',
            name_prop='name',
            query_props=query_props,
            case_insensitive_names=True)

    @property
    def console(self):
        """
        :class:`~zhmcclient.Console`: :term:`Console` defining the scope for
        this manager.
        """
        return self._parent

    @logged_api_call
    def list(self, full_properties=False, filter_args=None):
        """
        List the :term:`User` resources representing the users defined in this
        HMC.

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way:

        * If this manager is enabled for :ref:`auto-updating`, a locally
          maintained resource list is used (which is automatically updated via
          inventory notifications from the HMC) and the provided filter
          arguments are applied.

        * Otherwise, if the filter arguments specify the resource name as a
          single filter argument with a straight match string (i.e. without
          regular expressions), an optimized lookup is performed based on a
          locally maintained name-URI cache.

        * Otherwise, the HMC List operation is performed with the subset of the
          provided filter arguments that can be handled on the HMC side and the
          remaining filter arguments are applied on the client side on the list
          result.

        HMC/SE version requirements:

        * HMC version >= 2.13.0

        Authorization requirements:

        * User-related-access permission to the User object included in the
          result, or, depending on the type of User object, task permission to
          the "Manage Users" task or the "Manage User Templates" task.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only the short set as returned by the list
            operation.

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all resources are
            returned.

        Returns:

          : A list of :class:`~zhmcclient.User` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result_prop = 'users'
        list_uri = f'{self.console.uri}/users'
        return self._list_with_operation(
            list_uri, result_prop, full_properties, filter_args, None)

    @logged_api_call(blanked_properties=['password'], properties_pos=1)
    def create(self, properties):
        """
        Create a new User in this HMC.

        HMC/SE version requirements:

        * HMC version >= 2.13.0

        Authorization requirements:

        * Task permission to the "Manage Users" task to create a standard user
          or the "Manage User Templates" task to create a template user.

        Parameters:

          properties (dict): Initial property values.
            Allowable properties are defined in section 'Request body contents'
            in section 'Create User' in the :term:`HMC API` book.

        Returns:

          User:
            The resource object for the new User.
            The object will have its 'object-uri' property set as returned by
            the HMC, and will also have the input properties set.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result = self.session.post(self.console.uri + '/users',
                                   body=properties)
        # There should not be overlaps, but just in case there are, the
        # returned props should overwrite the input props:
        props = copy.deepcopy(properties)
        props.update(result)
        props.pop('password', None)
        name = props.get(self._name_prop, None)
        uri = props[self._uri_prop]
        user = User(self, uri, name, props)
        self._name_uri_cache.update(name, uri)
        return user


class User(BaseResource):
    """
    Representation of a :term:`User`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.UserManager`).

    HMC/SE version requirements:

    * HMC version >= 2.13.0
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.UserManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        assert isinstance(manager, UserManager), (
            f"Console init: Expected manager type {UserManager}, "
            f"got {type(manager)}")
        super().__init__(manager, uri, name, properties)

    @logged_api_call
    def delete(self):
        """
        Delete this User.

        HMC/SE version requirements:

        * HMC version >= 2.13.0

        Authorization requirements:

        * Task permission to the "Manage Users" task to delete a non-template
          user, or the "Manage User Templates" task to delete a template user.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.delete(self.uri, resource=self)
        self.manager._name_uri_cache.delete(
            self.get_properties_local(self.manager._name_prop, None))
        self.cease_existence_local()

    @logged_api_call(blanked_properties=['password'], properties_pos=1)
    def update_properties(self, properties):
        """
        Update writeable properties of this User.

        This method serializes with other methods that access or change
        properties on the same Python object.

        HMC/SE version requirements:

        * HMC version >= 2.13.0

        Authorization requirements:

        * Task permission to the "Manage Users" task to update a non-template
          user, or the "Manage User Templates" task to update a template user.
        * For a user to update their own password or default-group-uri
          property, user-related-access permission to the user represented
          by this User object, or task permission to the "Manage Users" task is
          required.

        Parameters:

          properties (dict): New values for the properties to be updated.
            Properties not to be updated are omitted.
            Allowable properties are the properties with qualifier (w) in
            section 'Data model' in section 'User object' in the
            :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.post(self.uri, resource=self, body=properties)

        # The name of Users cannot be updated. An attempt to do so should cause
        # HTTPError to be raised in the POST above, so we assert that here,
        # because we omit the extra code for handling name updates:
        assert self.manager._name_prop not in properties
        props = copy.deepcopy(properties)
        props.pop('password', None)
        self.update_properties_local(props)

    @logged_api_call
    def add_user_role(self, user_role):
        """
        Add the specified User Role to this User.

        This User must not be a system-defined or pattern-based user.

        HMC/SE version requirements:

        * HMC version >= 2.13.0

        Authorization requirements:

        * Task permission to the "Manage Users" task to modify a standard user
          or the "Manage User Templates" task to modify a template user.

        Parameters:

          user_role (:class:`~zhmcclient.UserRole`): User Role to be added.
            Must not be `None`.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {
            'user-role-uri': user_role.uri
        }
        self.manager.session.post(
            self.uri + '/operations/add-user-role', resource=self,
            body=body)

    @logged_api_call
    def remove_user_role(self, user_role):
        """
        Remove the specified User Role from this User.

        This User must not be a system-defined or pattern-based user.

        HMC/SE version requirements:

        * HMC version >= 2.13.0

        Authorization requirements:

        * Task permission to the "Manage Users" task to modify a standard user
          or the "Manage User Templates" task to modify a template user.

        Parameters:

          user_role (:class:`~zhmcclient.UserRole`): User Role to be removed.
            Must not be `None`.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {
            'user-role-uri': user_role.uri
        }
        self.manager.session.post(
            self.uri + '/operations/remove-user-role', resource=self,
            body=body)



# Copyright 2017,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The HMC supports the retrieval of metrics values for resources of a IBM Z
or LinuxONE computer. This section describes the zhmcclient API for retrieving
such metrics from the HMC.

A resource termed :term:`Metrics Context` is associated with any metrics
retrieval. These resources are user-created definitions of the kinds of metrics
that are intended to be retrieved. A metrics context mostly defines the names
of the metric groups to be retrieved. The available metric groups are described
in section 'Metric Groups' in the :term:`HMC API` book.

The zhmcclient API for metrics provides access to the metric values and to
their definitions, so that clients using the metric values do not need to have
intimate knowledge about the specific metric values when just displaying them.

The basic usage of the metrics API is shown in this example:

.. code-block:: python

    # Create a Metrics Context for the desired metric groups:
    metric_groups = ['dpm-system-usage-overview', 'partition-usage']
    mc = client.metrics_contexts.create(
        {'anticipated-frequency-seconds': 15,
         'metric-groups': metric_groups})

    # Retrieve the current metric values:
    mr_str = mc.get_metrics()

    # Display the metric values:
    print("Current metric values:")
    mr = zhmcclient.MetricsResponse(mc, mr_str)
    for mg in mr.metric_group_values:
        mg_name = mg.name
        mg_def = mc.metric_group_definitions[mg_name]
        print(f"  Metric group: {mg_name}")
        for ov in mg.object_values:
            print(f"    Resource: {ov.resource_uri}")
            print(f"    Timestamp: {ov.timestamp}")
            print("    Metric values:")
            for m_name in ov.metrics:
                m_value = ov.metrics[m_name]
                m_def = mg_def.metric_definitions[m_name]
                m_unit = m_def.unit or ''
                print(f"      {m_name:30}  {m_value} {m_unit}")
        if not mg.object_values:
            print("    No resources")

    # Delete the Metrics Context:
    mc.delete()
"""


from collections import OrderedDict, namedtuple
import re
from datetime import datetime
import pytz

from ._manager import BaseManager
from ._resource import BaseResource
from ._cpc import Cpc
from ._lpar import Lpar
from ._partition import Partition
from ._adapter import Adapter
from ._nic import Nic
from ._logging import logged_api_call
from ._exceptions import NotFound, MetricsResourceNotFound
from ._utils import datetime_from_timestamp, repr_list, datetime_to_isoformat, \
    repr_obj_id

__all__ = ['MetricsContextManager', 'MetricsContext', 'MetricGroupDefinition',
           'MetricDefinition', 'MetricsResponse', 'MetricGroupValues',
           'MetricObjectValues']


class MetricsContextManager(BaseManager):
    """
    Manager providing access to the :term:`Metrics Context` resources that
    were created through this manager object.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible via the :attr:`~zhmcclient.Client.metrics_contexts` attribute
    of the :class:`~zhmcclient.Client` object connected to the HMC.

    HMC/SE version requirements: None
    """

    def __init__(self, client):
        # This function should not go into the docs.
        # Parameters:
        #   client (:class:`~zhmcclient.Client`):
        #      Client object for the HMC to be used.
        super().__init__(
            resource_class=MetricsContext,
            class_name='',
            session=client.session,
            parent=None,
            base_uri='/api/services/metrics/context',
            oid_prop='',
            uri_prop='metrics-context-uri',
            name_prop='',
            query_props=[])

        self._client = client
        self._metrics_contexts = []

    def __repr__(self):
        """
        Return a string with the state of this manager object, for debug
        purposes.
        """
        ret = (
            f"{repr_obj_id(self)} (\n"
            f"  _resource_class={self._resource_class!r},\n"
            f"  _class_name={self._class_name!r},\n"
            f"  _session={repr_obj_id(self._session)},\n"
            f"  _parent={repr_obj_id(self._parent)},\n"
            f"  _base_uri={self._base_uri!r},\n"
            f"  _oid_prop={self._oid_prop!r},\n"
            f"  _uri_prop={self._uri_prop!r},\n"
            f"  _name_prop={self._name_prop!r},\n"
            f"  _query_props={repr_list(self._query_props, indent=2)},\n"
            f"  _list_has_name={self._list_has_name!r},\n"
            f"  _name_uri_cache={self._name_uri_cache!r},\n"
            f"  _client={repr_obj_id(self._client)},\n"
            "  _metrics_contexts="
            f"{repr_list(self._metrics_contexts, indent=2)},\n"
            ")")
        return ret

    @property
    def client(self):
        """
        :class:`~zhmcclient.Client`:
          The client defining the scope for this manager.
        """
        return self._client

    @logged_api_call
    def list(self, full_properties=False):
        # pylint: disable=arguments-differ
        """
        List the :term:`Metrics Context` resources that were created through
        this manager object.

        Note that the HMC does not provide a way to enumerate the existing
        :term:`Metrics Context` resources. Therefore, this method will only
        list the :term:`Metrics Context` resources that were created through
        this manager object. For example, :term:`Metrics Context` resources
        created through a second :class:`~zhmcclient.Client` object will not be
        listed.

        HMC/SE version requirements: None

        Parameters:

          full_properties (bool):
            This parameter exists for compatibility with other resource
            classes, but for this class, it has no effect on the result.

        Returns:

          : A list of :class:`~zhmcclient.MetricsContext` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        return self._metrics_contexts

    @logged_api_call
    def create(self, properties):
        """
        Create a :term:`Metrics Context` resource in the HMC this client is
        connected to.

        HMC/SE version requirements: None

        Parameters:

          properties (dict): Initial property values.
            Allowable properties are defined in section 'Request body contents'
            in section 'Create Metrics Context' in the :term:`HMC API` book.

        Returns:

          :class:`~zhmcclient.MetricsContext`:
            The resource object for the new :term:`Metrics Context` resource.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result = self.session.post('/api/services/metrics/context',
                                   body=properties)
        mc_properties = properties.copy()
        mc_properties.update(result)
        new_metrics_context = MetricsContext(self,
                                             result['metrics-context-uri'],
                                             None,
                                             mc_properties)
        self._metrics_contexts.append(new_metrics_context)
        return new_metrics_context


class MetricsContext(BaseResource):
    """
    Representation of a :term:`Metrics Context` resource.

    A :term:`Metrics Context` resource specifies a list of metrics groups for
    which the current metric values can be retrieved using the
    :meth:`~zhmcclient.MetricsContext.get_metrics` method.

    The properties of this resource are the response fields described for the
    'Create Metrics Context' operation in the :term:`HMC API` book.

    This class is derived from :class:`~zhmcclient.BaseResource`; see there
    for common methods and attributes.

    Objects of this class can be created by the user with the
    :meth:`zhmcclient.MetricsContextManager.create` method.
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.MetricsContextManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        if not isinstance(manager, MetricsContextManager):
            raise AssertionError(
                "MetricsContext init: Expected manager type "
                f"{MetricsContextManager}, got {type(manager)}")
        super().__init__(manager, uri, name, properties)

        self._metric_group_definitions = self._setup_metric_group_definitions()

    def _setup_metric_group_definitions(self):
        """
        Return the dict of MetricGroupDefinition objects for this metrics
        context, by processing its 'metric-group-infos' property.
        """
        # Dictionary of MetricGroupDefinition objects, by metric group name
        metric_group_definitions = {}
        for mg_info in self._properties['metric-group-infos']:
            mg_name = mg_info['group-name']
            mg_def = MetricGroupDefinition(
                name=mg_name,
                resource_class=_resource_class_from_group(mg_name),
                metric_definitions={})
            for i, m_info in enumerate(mg_info['metric-infos']):
                m_name = m_info['metric-name']
                m_def = MetricDefinition(
                    index=i,
                    name=m_name,
                    type=_metric_type(m_info['metric-type']),
                    unit=_metric_unit_from_name(m_name))
                mg_def.metric_definitions[m_name] = m_def
            metric_group_definitions[mg_name] = mg_def
        return metric_group_definitions

    @property
    def metric_group_definitions(self):
        """
        dict: The metric definitions for the metric groups of this
          :term:`Metrics Context` resource, as a dictionary of
          :class:`~zhmcclient.MetricGroupDefinition` objects, by metric group
          name.
        """
        return self._metric_group_definitions

    @logged_api_call
    def get_metrics(self):
        """
        Retrieve the current metric values for this :term:`Metrics Context`
        resource from the HMC.

        The metric values are returned by this method as a string in the
        `MetricsResponse` format described with the 'Get Metrics' operation in
        the :term:`HMC API` book.

        The :class:`~zhmcclient.MetricsResponse` class can be used to process
        the `MetricsResponse` string returned by this method, and provides
        structured access to the metrics values.

        HMC/SE version requirements: None

        Returns:

          :term:`string`:
            The current metric values, in the `MetricsResponse` string format.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        metrics_response = self.manager.session.get(self.uri, resource=self)
        return metrics_response

    @logged_api_call
    def delete(self):
        """
        Delete this :term:`Metrics Context` resource.

        HMC/SE version requirements: None

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.delete(self.uri, resource=self)
        self.manager._metrics_contexts.remove(self)


_MetricGroupDefinitionTuple = namedtuple(
    '_MetricGroupDefinitionTuple',
    ['name', 'resource_class', 'metric_definitions']
)


class MetricGroupDefinition(_MetricGroupDefinitionTuple):
    """
    A :func:`namedtuple <py:collections.namedtuple>` representing definitional
    information for a metric group.

    HMC/SE version requirements: None
    """

    def __new__(cls, name, resource_class, metric_definitions):
        """
        Parameters:

          name (:term:`string`):
            Metric group name, as defined in section 'Metric groups' in the
            :term:`HMC API` book.

          resource_class (:term:`string`):
            A string identifying the resource class to which this metric group
            belongs, using the values from the 'class' property of resource
            objects.

          metric_definitions (dict):
            Metric definitions for the metrics in this metric group, as a
            dictionary where the key is the metric name and the value is the
            :class:`~zhmcclient.MetricDefinition` object for the metric.

        All these parameters are also available as same-named attributes.
        """
        self = super().__new__(
            cls, name, resource_class, metric_definitions)
        return self

    __slots__ = ()

    def __repr__(self):
        repr_str = (
            "MetricGroupDefinition("
            f"name={self.name!r}, "
            f"resource_class={self.resource_class!r}, "
            f"metric_definitions={self.metric_definitions!r})")
        return repr_str


_MetricDefinitionTuple = namedtuple(
    '_MetricDefinitionTuple',
    ['index', 'name', 'type', 'unit']
)


class MetricDefinition(_MetricDefinitionTuple):
    """
    A :func:`namedtuple <py:collections.namedtuple>` representing definitional
    information for a single metric.

    HMC/SE version requirements: None
    """

    def __new__(cls, index, name, type, unit):
        # pylint: disable=redefined-builtin
        """
        Parameters:

          index (:term:`integer`):
            0-based index (=position) of the metric in a MetricsResponse value
            row.

          name (:term:`string`):
            Metric field name, as shown in the tables defining the metric
            groups in section 'Metric groups' in the :term:`HMC API` book.

          type (:term:`callable`):
            Python type for the metric value. The type must be a constructor
            (callable) that takes the metrics value from the `MetricsResponse`
            string as its only argument, using the following Python types
            for the metric group description types shown in the :term:`HMC API`
            book:

            =============================  ======================
            Description type               Python type
            =============================  ======================
            Boolean                        :class:`py:bool`
            Byte                           :term:`integer`
            Short                          :term:`integer`
            Integer                        :term:`integer`
            Long                           :term:`integer`
            Double                         :class:`py:float`
            String, String Enum            :term:`unicode string`
            =============================  ======================

          unit (:term:`string`):
            Unit of the metric value.

        All these parameters are also available as same-named attributes.
        """
        self = super().__new__(
            cls, index, name, type, unit)
        return self

    __slots__ = ()

    def __repr__(self):
        repr_str = (
            "MetricDefinition("
            f"index={self.index!r}, "
            f"name={self.name!r}, "
            f"type={self.type!r}, "
            f"unit={self.unit!r})")
        return repr_str


def _metric_type(metric_type_name):
    """
    Return a constructor callable for the given metric type name.

    The returned callable takes the metric value as a string as its only
    argument and returns a Python object representing that metric value using
    the correct Python type.
    """
    return _METRIC_TYPES_BY_NAME[metric_type_name]


_METRIC_TYPES_BY_NAME = {
    'boolean-metric': bool,
    'byte-metric': int,
    'short-metric': int,
    'integer-metric': int,
    'long-metric': int,
    'double-metric': float,
    'string-metric': str,
}


def _metric_value(value_str, metric_type):
    """
    Return a Python-typed metric value from a metric value string.
    """
    if metric_type in (int, float):
        try:
            return metric_type(value_str)
        except ValueError:
            new_exc = ValueError(
                f"Invalid {metric_type.__class__.__name__} metric value: "
                f"{value_str!r}")
            new_exc.__cause__ = None
            raise new_exc  # ValueError
    elif metric_type is str:
        # In Python 3, decode('unicode_escape) requires bytes, so we need
        # to encode to bytes. This also works in Python 2.
        return value_str.strip('"').encode('utf-8').decode('unicode_escape')
    else:
        assert metric_type is bool
        lower_str = value_str.lower()
        if lower_str == 'true':
            return True
        if lower_str == 'false':
            return False
        raise ValueError(f"Invalid boolean metric value: {value_str!r}")


def _metric_unit_from_name(metric_name):
    """
    Return a metric unit string for human consumption, that is inferred from
    the metric name.

    If a unit cannot be inferred, `None` is returned.
    """
    for item in _PATTERN_UNIT_LIST:
        pattern, unit = item
        if pattern.match(metric_name):
            return unit
    return None


_USE_UNICODE = True
if _USE_UNICODE:
    MICROSECONDS = "\u00b5s"  # U+00B5 = Micro Sign
    CELSIUS = "\u00B0C"  # U+00B0 = Degree Sign
    # Note: Use of U+2103 (Degree Celsius) is discouraged by Unicode standard
else:
    MICROSECONDS = "us"
    CELSIUS = "degree Celsius"  # Official SI unit when not using degree sign


_PATTERN_UNIT_LIST = {
    # End patterns:
    (re.compile(r".+-usage$"), "%"),
    (re.compile(r".+-time$"), MICROSECONDS),
    (re.compile(r".+-time-used$"), MICROSECONDS),
    (re.compile(r".+-celsius$"), CELSIUS),
    (re.compile(r".+-watts$"), "W"),
    (re.compile(r".+-paging-rate$"), "pages/s"),
    (re.compile(r".+-sampling-rate$"), "samples/s"),
    # Begin patterns:
    (re.compile(r"^bytes-.+"), "B"),
    (re.compile(r"^heat-load.+"), "BTU/h"),  # Note: No trailing hyphen
    (re.compile(r"^interval-bytes-.+"), "B"),
    (re.compile(r"^bytes-per-second-.+"), "B/s"),
    # Special cases:
    (re.compile(r"^storage-rate$"), "kB/s"),
    (re.compile(r"^humidity$"), "%"),
    (re.compile(r"^memory-used$"), "MiB"),
    (re.compile(r"^policy-activation-time$"), ""),  # timestamp
    (re.compile(r"^velocity-numerator$"), MICROSECONDS),
    (re.compile(r"^velocity-denominator$"), MICROSECONDS),
    (re.compile(r"^utilization$"), "%"),
}


def _resource_class_from_group(metric_group_name):
    """
    Return the resource class string from the metric group name.

    Metric groups for resources that are specific to ensemble mode are not
    supported.

    Returns an empty string if a metric group name is unknown.
    """
    return CLASS_FROM_GROUP.get(metric_group_name, '')


CLASS_FROM_GROUP = {
    # DPM mode only:
    'dpm-system-usage-overview': 'cpc',
    'partition-usage': 'partition',
    'adapter-usage': 'adapter',
    'network-physical-adapter-port': 'adapter',
    'partition-attached-network-interface': 'nic',
    # Classic mode only:
    'cpc-usage-overview': 'cpc',
    'logical-partition-usage': 'logical-partition',
    'channel-usage': 'cpc',
    'crypto-usage': 'cpc',
    'flash-memory-usage': 'cpc',  # TODO: verify CPC mode dependency
    'roce-usage': 'cpc',  # TODO: verify CPC mode dependency
    # DPM mode or classic mode:
    'zcpc-environmentals-and-power': 'cpc',
    'zcpc-processor-usage': 'cpc',
}


class MetricsResponse:
    """
    Represents the metric values returned by one call to the
    :meth:`~zhmcclient.MetricsContext.get_metrics` method, and provides
    structured access to the data.

    HMC/SE version requirements: None
    """

    def __init__(self, metrics_context, metrics_response_str):
        """
        Parameters:

          metrics_context (:class:`~zhmcclient.MetricsContext`):
            The :class:`~zhmcclient.MetricsContext` object that was used to
            retrieve the metrics response string. It defines the structure of
            the metric values in the metrics response string.

          metrics_response_str (:term:`string`):
            The metrics response string, as returned by the
            :meth:`~zhmcclient.MetricsContext.get_metrics` method.
        """
        self._metrics_context = metrics_context
        self._metrics_response_str = metrics_response_str
        self._client = self._metrics_context.manager.client

        self._metric_group_values = self._setup_metric_group_values()

    def _setup_metric_group_values(self):
        """
        Return the list of MetricGroupValues objects for this metrics response,
        by processing its metrics response string.

        The lines in the metrics response string are::

            MetricsResponse: MetricsGroup{0,*}
                             <emptyline>      a third empty line at the end

            MetricsGroup:    MetricsGroupName
                             ObjectValues{0,*}
                             <emptyline>      a second empty line after each MG

            ObjectValues:    ObjectURI
                             Timestamp
                             ValueRow{1,*}
                             <emptyline>      a first empty line after this blk
        """

        mg_defs = self._metrics_context.metric_group_definitions

        metric_group_name = None
        resource_uri = None
        dt_timestamp = None

        object_values = None
        metric_group_values = []
        state = 0
        for mr_line in self._metrics_response_str.splitlines():
            if state == 0:
                if object_values is not None:
                    # Store the result from the previous metric group
                    mgv = MetricGroupValues(metric_group_name, object_values)
                    metric_group_values.append(mgv)
                    object_values = None
                if mr_line == '':
                    # Skip initial (or trailing) empty lines
                    pass
                else:
                    # Process the next metrics group
                    metric_group_name = mr_line.strip('"')  # No " or \ inside
                    assert metric_group_name in mg_defs
                    m_defs = mg_defs[metric_group_name].metric_definitions
                    object_values = []
                    state = 1
            elif state == 1:
                if mr_line == '':
                    # There are no (or no more) ObjectValues items in this
                    # metrics group
                    state = 0
                else:
                    # There are ObjectValues items
                    resource_uri = mr_line.strip('"')  # No " or \ inside
                    state = 2
            elif state == 2:
                # Process the timestamp
                assert mr_line != ''
                try:
                    dt_timestamp = datetime_from_timestamp(int(mr_line))
                except ValueError:
                    # Sometimes, the returned epoch timestamp values are way
                    # too large, e.g. 3651584404810066 (which would translate
                    # to the year 115791 A.D.). Python datetime supports
                    # up to the year 9999. We circumvent this issue by
                    # simply using the current date&time.
                    # TODO: Remove the circumvention for too large timestamps.
                    dt_timestamp = datetime.now(pytz.utc)
                state = 3
            elif state == 3:
                if mr_line != '':
                    # Process the metric values in the ValueRow line
                    str_values = mr_line.split(',')
                    metrics = {}
                    # pylint: disable=possibly-used-before-assignment
                    for m_name in m_defs:
                        m_def = m_defs[m_name]
                        m_type = m_def.type
                        m_value_str = str_values[m_def.index]
                        m_value = _metric_value(m_value_str, m_type)
                        metrics[m_name] = m_value
                    ov = MetricObjectValues(
                        self._client, mg_defs[metric_group_name], resource_uri,
                        dt_timestamp, metrics)
                    object_values.append(ov)
                    # stay in this state, for more ValueRow lines
                else:
                    # On the empty line after the last ValueRow line
                    state = 1

        return metric_group_values

    @property
    def metrics_context(self):
        """
        :class:`~zhmcclient.MetricsContext` object for this metric response.
        This can be used to access the metric definitions for this response.
        """
        return self._metrics_context

    @property
    def metric_group_values(self):
        """
        :class:`py:list`: The list of :class:`~zhmcclient.MetricGroupValues`
          objects representing the metric groups in this metric response.

          Each :class:`~zhmcclient.MetricGroupValues` object contains a list of
          :class:`~zhmcclient.MetricObjectValues` objects representing the
          metric values in this group (each for a single resource and point in
          time).
        """
        return self._metric_group_values


class MetricGroupValues:
    """
    Represents the metric values for a metric group in a MetricsResponse
    string.

    HMC/SE version requirements: None
    """

    def __init__(self, name, object_values):
        """
        Parameters:

          name (:term:`string`):
            Metric group name.

          object_values (:class:`py:list`):
            The :class:`~zhmcclient.MetricObjectValues` objects in this metric
            group. Each of them represents the metric values for a single
            resource at a single point in time.
        """
        self._name = name
        self._object_values = object_values

    @property
    def name(self):
        """
        string: The metric group name.
        """
        return self._name

    @property
    def object_values(self):
        """
        :class:`py:list`: The :class:`~zhmcclient.MetricObjectValues` objects
          in this metric group. Each of them represents the metric values for
          a single resource at a single point in time.
        """
        return self._object_values


class MetricObjectValues:
    """
    Represents the metric values for a single resource at a single point in
    time.

    HMC/SE version requirements: None
    """

    def __init__(self, client, metric_group_definition, resource_uri,
                 timestamp, metrics):
        """
        Parameters:

          client (:class:`~zhmcclient.Client`):
            Client object, for retrieving the actual resource.

          metric_group_definition (:class:`~zhmcclient.MetricGroupDefinition`):
            Metric group definition for this set of metric values.

          resource_uri (:term:`string`):
            Resource URI of the resource these metric values apply to.

          timestamp (:class:`py:datetime.datetime`):
            Point in time when the HMC captured these metric values (as a
            timezone-aware datetime object).

          metrics (dict):
            The metric values, as a dictionary of the (Python typed) metric
            values, by metric name.
        """
        self._client = client
        self._metric_group_definition = metric_group_definition
        self._resource_uri = resource_uri
        self._timestamp = timestamp
        self._metrics = metrics
        self._resource = None  # Lazy initialization

    @property
    def client(self):
        """
        :class:`~zhmcclient.Client`: Client object, for retrieving the actual
        resource.
        """
        return self._client

    @property
    def metric_group_definition(self):
        """
        :class:`~zhmcclient.MetricGroupDefinition`: Metric group definition for
        this set of metric values.
        """
        return self._metric_group_definition

    @property
    def resource_uri(self):
        """
        string: The canonical URI path of the resource these metric values
        apply to.

        Example: ``/api/cpcs/12345``
        """
        return self._resource_uri

    @property
    def timestamp(self):
        """
        :class:`py:datetime.datetime`: Point in time when the HMC captured
          these metric values (as a timezone-aware datetime object).
        """
        return self._timestamp

    @property
    def metrics(self):
        """
        dict: The metric values, as a dictionary of the (Python typed) metric
          values, by metric name.
        """
        return self._metrics

    @property
    def resource(self):
        """
        :class:`~zhmcclient.BaseResource`: The Python resource object of the
          resource these metric values apply to.

        Raises:

          :exc:`~zhmcclient.MetricsResourceNotFound`: The resource for the
          resource URI of the resource these metric values apply to, was not
          found on the HMC.
        """
        if self._resource is not None:
            return self._resource

        resource_class = self.metric_group_definition.resource_class
        resource_uri = self.resource_uri

        if resource_class == 'cpc':
            cpc_managers = [self.client.cpcs]
            filter_args = {'object-uri': resource_uri}
            try:
                resource = self.client.cpcs.find(**filter_args)
            except NotFound:
                raise MetricsResourceNotFound(
                    f"{resource_class} with URI {resource_uri} not found on "
                    f"HMC {self.client.session.host}",
                    Cpc, cpc_managers)
        elif resource_class == 'logical-partition':
            lpar_managers = []
            cpc_list = self.client.cpcs.list()
            for cpc in cpc_list:
                lpar_managers.append(cpc.lpars)
                try:
                    filter_args = {'object-uri': resource_uri}
                    resource = cpc.lpars.find(**filter_args)
                    break
                except NotFound:
                    pass  # Try next CPC
            else:
                raise MetricsResourceNotFound(
                    f"{resource_class} with URI {resource_uri} not found in "
                    f"CPCs {', '.join([cpc.name for cpc in cpc_list])}",
                    Lpar, lpar_managers)
        elif resource_class == 'partition':
            partitions = self.client.consoles.console.list_permitted_partitions(
                filter_args={'object-uri': resource_uri})
            if len(partitions) < 1:
                raise MetricsResourceNotFound(
                    f"Partition with URI {resource_uri} not found",
                    Partition, [])
            resource = partitions[0]
        elif resource_class == 'adapter':
            cpc_list = self.client.cpcs.list()
            adapter_managers = []
            for cpc in cpc_list:
                adapter_managers.append(cpc.adapters)
                try:
                    filter_args = {'object-uri': resource_uri}
                    resource = cpc.adapters.find(**filter_args)
                    break
                except NotFound:
                    pass  # Try next CPC
            else:
                raise MetricsResourceNotFound(
                    f"{resource_class} with URI {resource_uri} not found in "
                    f"CPCs {', '.join([cpc.name for cpc in cpc_list])}",
                    Adapter, adapter_managers)
        elif resource_class == 'nic':
            nic_properties = self.client.session.get(resource_uri)
            partition_uri = nic_properties['parent']
            partitions = self.client.consoles.console.list_permitted_partitions(
                filter_args={'object-uri': partition_uri})
            if len(partitions) < 1:
                raise MetricsResourceNotFound(
                    f"Parent partition with URI {partition_uri} of NIC with "
                    f"URI {resource_uri} not found",
                    Partition, [])
            partition = partitions[0]
            nic_managers = [partition.nics]
            filter_args = {'element-uri': resource_uri}
            try:
                resource = partition.nics.find(**filter_args)
            except NotFound:
                raise MetricsResourceNotFound(
                    f"NIC with URI {resource_uri} not found in its parent "
                    f"partition {cpc.name}.{partition.name}",
                    Nic, nic_managers)
        else:
            raise ValueError(
                f"Invalid resource class: {resource_class!r}")

        self._resource = resource
        return self._resource

    def dump(self):
        """
        Dump these metric values as a resource definition.

        The timestamp of the object is represented as an ISO8601 string using
        this format::

            YYYY-MM-DD HH:MM:SS[.ssssss]shh:mm

        Where:

          * YYYY-MM-DD - is year, month and day.

          * HH:MM:SS - is hour in 24-hour format, minutes and seconds.

          * .ssssss - is an optional part specifying microseconds. It is not
            created when the timestamp microsecond value is 0.

          * shh:mm - is the timezone offset from UTC with sign, hours (hh) and
            minutes (mm). Since the FakedMetricObjectValues class ensures that
            the timestamp is always timezone-aware, this part is always created.

        Returns:

          dict: Resource definition of this object.
        """
        obj_dict = OrderedDict()
        # Faked Simple properties
        obj_dict['group_name'] = self.metric_group_definition.name
        obj_dict['resource_uri'] = self.resource_uri
        obj_dict['timestamp'] = datetime_to_isoformat(self.timestamp)
        obj_dict['metrics'] = self.metrics
        return obj_dict



# Copyright 2017,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A :term:`CPC` (Central Processor Complex) is a physical IBM Z or LinuxONE
computer.

A particular HMC can manage multiple CPCs and can discover other CPCs that
are not managed by that HMC. Such other CPCs are called "unmanaged CPCs" and
they may or may not be managed by another HMC.

This section describes the interface for *unmanaged* CPCs using resource class
:class:`~zhmcclient.UnmanagedCpc` and the corresponding manager class
:class:`~zhmcclient.UnmanagedCpcManager`.
"""


from ._manager import BaseManager
from ._resource import BaseResource
from ._logging import logged_api_call
from ._utils import RC_CPC

__all__ = ['UnmanagedCpcManager', 'UnmanagedCpc']


class UnmanagedCpcManager(BaseManager):
    """
    Manager providing access to the :term:`CPCs <CPC>` that have been
    discovered by the HMC this client is connected to, but are not managed by
    it. They may or may not be managed by another HMC.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible via the following instance variable of a
    :class:`~zhmcclient.Console` object:

    * :attr:`~zhmcclient.Console.unmanaged_cpcs`

    HMC/SE version requirements:

    * HMC version >= 2.13.1
    """

    def __init__(self, console):
        # This function should not go into the docs.
        # Parameters:
        #   console (:class:`~zhmcclient.Console`):
        #      Console object for the HMC to be used.

        # Resource properties that are supported as filter query parameters
        # (for server-side filtering).
        query_props = [
            'name',
        ]

        super().__init__(
            resource_class=UnmanagedCpc,
            class_name=RC_CPC,
            session=console.manager.session,
            parent=console,
            base_uri='/api/console',
            oid_prop='object-id',
            uri_prop='object-uri',
            name_prop='name',
            query_props=query_props)

    @property
    def console(self):
        """
        :class:`~zhmcclient.Console`: :term:`Console` defining the scope for
        this manager.
        """
        return self._parent

    @logged_api_call
    def list(self, full_properties=False, filter_args=None):
        """
        List the unmanaged CPCs exposed by the HMC this client is connected to.

        Because the CPCs are unmanaged, the returned
        :class:`~zhmcclient.UnmanagedCpc` objects cannot perform any operations
        and will have only the following properties:

        * ``object-uri``
        * ``name``

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way:

        * If this manager is enabled for :ref:`auto-updating`, a locally
          maintained resource list is used (which is automatically updated via
          inventory notifications from the HMC) and the provided filter
          arguments are applied.

        * Otherwise, if the filter arguments specify the resource name as a
          single filter argument with a straight match string (i.e. without
          regular expressions), an optimized lookup is performed based on a
          locally maintained name-URI cache.

        * Otherwise, the HMC List operation is performed with the subset of the
          provided filter arguments that can be handled on the HMC side and the
          remaining filter arguments are applied on the client side on the list
          result.

        HMC/SE version requirements:

        * HMC version >= 2.13.1

        Authorization requirements: None

        Parameters:

          full_properties (bool):
            Ignored (exists for consistency with other list() methods).

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all resources are
            returned.

        Returns:

          : A list of :class:`~zhmcclient.UnmanagedCpc` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result_prop = 'cpcs'
        list_uri = f'{self.parent.uri}/operations/list-unmanaged-cpcs'
        return self._list_with_operation(
            list_uri, result_prop, full_properties, filter_args, None)


class UnmanagedCpc(BaseResource):
    """
    Representation of an unmanaged :term:`CPC`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.UnmanagedCpcManager`).

    HMC/SE version requirements:

    * HMC version >= 2.13.1
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.UnmanagedCpcManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        assert isinstance(manager, UnmanagedCpcManager), (
            f"UnmanagedCpc init: Expected manager type {UnmanagedCpcManager}, "
            f"got {type(manager)}")
        super().__init__(manager, uri, name, properties)



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The zhmcclient supports logging using the standard Python :mod:`py:logging`
module, using standard Python :class:`~py:logging.Logger` objects with these
names:

* 'zhmcclient.api' for user-issued calls to zhmcclient API functions, at the
  debug level. Internal calls to API functions are not logged.

* 'zhmcclient.hmc' for operations from zhmcclient to the HMC, at the
  debug level.

* 'zhmcclient.jms' for notifications from the HMC to zhmcclient, at the
  debug, info, warning and error level. At this point, this logger is used only
  for the :ref:`auto-updating` support, but not for
  the :class:`~zhmcclient.NotificationReceiver` class.

* 'zhmcclient.os' for interactions with OS consoles, at the debug level.

For HMC operations and API calls that contain the HMC password or HMC session
tokens, the password is hidden in the log message by replacing it with a few
'*' characters.

All these loggers have a null-handler (see :class:`~py:logging.NullHandler`)
and have no log formatter (see :class:`~py:logging.Formatter`).

As a result, the loggers are silent by default. If you want to turn on logging,
add a log handler (see :meth:`~py:logging.Logger.addHandler`, and
:mod:`py:logging.handlers` for the handlers included with Python) and set the
log level (see :meth:`~py:logging.Logger.setLevel`, and :ref:`py:levels` for
the defined levels).

If you want to change the default log message format, use
:meth:`~py:logging.Handler.setFormatter`. Its ``form`` parameter is a format
string with %-style placeholders for the log record attributes (see Python
section :ref:`py:logrecord-attributes`).

Examples:

* To output the log records for all HMC operations to ``stdout`` in a
  particular format, do this::

      import logging

      handler = logging.StreamHandler()
      format_string = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
      handler.setFormatter(logging.Formatter(format_string))
      logger = logging.getLogger('zhmcclient.hmc')
      logger.addHandler(handler)
      logger.setLevel(logging.DEBUG)

* This example uses the :func:`~py:logging.basicConfig` convenience function
  that sets the same format and level as in the previous example, but for the
  root logger. Therefore, it will output all log records, not just from this
  package::

      import logging

      format_string = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
      logging.basicConfig(format=format_string, level=logging.DEBUG)
"""

import logging
import inspect
import functools
from collections.abc import Mapping, Sequence

from ._constants import API_LOGGER_NAME, BLANKED_OUT_STRING

__all__ = []


def log_escaped(string):
    """
    Return the escaped input string, for use in log messages.
    """
    return string.replace('\n', ' ').replace('  ', ' ').replace('  ', ' ').\
        replace('  ', ' ')


def get_logger(name):
    """
    Return a :class:`~py:logging.Logger` object with the specified name.

    A :class:`~py:logging.NullHandler` handler is added to the logger if it
    does not have any handlers yet and if it is not the Python root logger.
    This prevents the propagation of log requests up the Python logger
    hierarchy, and therefore causes this package to be silent by default.
    """
    logger = logging.getLogger(name)
    if name != '' and not logger.handlers:
        logger.addHandler(logging.NullHandler())
    return logger


def logged_api_call(
        org_func=None, *, blanked_properties=None, properties_pos=None):
    """
    Function decorator that causes the decorated API function or method to log
    calls to itself to a logger.

    The logger's name is the dotted module name of the module defining the
    decorated function (e.g. 'zhmcclient._cpc').

    Parameters:

      org_func (function object): The original function being decorated.
        Will be `None` if the decorator is specified with its optional
        argument 'blanked_properties'.

      blanked_properties (list of str): Optional: List of properties in the
        'properties' argument of the decorated API function that should be
        blanked out before being logged. Can be used to hide password
        properties.
        This parameter is required when 'properties_pos' is used.
        This parameter must be specified as a keyword argument.

      properties_pos (int): Optional: 0-based index of the 'properties'
        parameter in the argument list of the decorated API function.
        For methods, the 'self' or 'cls' parameter is included in the position.
        This parameter is needed in case the properties are passed as a
        positional argument by the caller of the API function.
        This parameter is required when 'blanked_properties' is used.
        This parameter must be specified as a keyword argument.

    Returns:

      function object: The function wrappering the original function being
        decorated.

    Raises:

      TypeError: The @logged_api_call decorator must be used on a function or
        method (and not on top of the @property decorator).
    """

    if blanked_properties is not None and properties_pos is None:
        raise TypeError(
            "If the @logged_api_call decorator is specified with "
            "'blanked_properties', 'properties_pos' must also be specified.")

    if properties_pos is not None and blanked_properties is None:
        raise TypeError(
            "If the @logged_api_call decorator is specified with "
            "'properties_pos', 'blanked_properties' must also be specified.")

    if blanked_properties is not None and (
            not isinstance(blanked_properties, Sequence) or  # noqa: W504
            isinstance(blanked_properties, str)):
        raise TypeError(
            "The 'blanked_properties' parameter of the @logged_api_call "
            "decorator must be a list of strings.")

    def _decorate(func):
        """
        The actual decorator function that always gets the original decorated
        function, independent of whether the 'logged_api_call' decorator was
        specified with or without its optional arguments.

        Parameters:

          func (function object): The original function being decorated.
        """

        # Note that in this decorator function, we are in a module loading
        # context, where the decorated functions are being defined. When this
        # decorator function is called, its call stack represents the
        # definition of the decorated functions. Not all global definitions in
        # the module have been defined yet, and methods of classes that are
        # decorated with this decorator are still functions at this point (and
        # not yet methods).

        if not inspect.isfunction(func):
            raise TypeError("The @logged_api_call decorator must be used on a "
                            "function or method (and not on top of the "
                            "@property decorator)")

        try:
            # We avoid the use of inspect.getouterframes() because it is slow,
            # and use the pointers up the stack frame, instead.

            this_frame = inspect.currentframe()  # this function
            apifunc_frame = this_frame.f_back  # the decorated API function
            if org_func:
                # In this case, there is one more decorator function nesting
                apifunc_frame = apifunc_frame.f_back
            apifunc_owner = inspect.getframeinfo(apifunc_frame)[2]

        finally:
            # Recommended way to deal with frame objects to avoid ref cycles
            del this_frame
            del apifunc_frame

        # TODO: For inner functions, show all outer levels instead of just one.

        func_name = getattr(func, '__name__', '<unknown>')
        if apifunc_owner == '<module>':
            # The decorated API function is defined globally (at module level)
            apifunc_str = f'{func_name}()'
        else:
            # The decorated API function is defined in a class or in a function
            apifunc_str = f'{apifunc_owner}.{func_name}()'

        logger = get_logger(API_LOGGER_NAME)

        def is_external_call():
            """
            Return a boolean indicating whether the call to the decorated API
            function is made from outside of the zhmcclient package.
            """
            try:
                # We avoid the use of inspect.getouterframes() because it is
                # slow, and use the pointers up the stack frame, instead.

                this_frame = inspect.currentframe()  # this function
                log_api_call_frame = this_frame.f_back  # log_api_call()
                apifunc_frame = log_api_call_frame.f_back  # the decorated func
                apicaller_frame = apifunc_frame.f_back  # caller of API func
                apicaller_module = inspect.getmodule(apicaller_frame)
                if apicaller_module is None:
                    apicaller_module_name = "<unknown>"
                else:
                    apicaller_module_name = apicaller_module.__name__
            finally:
                # Recommended way to deal with frame objects to avoid ref
                # cycles
                del this_frame
                del log_api_call_frame
                del apifunc_frame
                del apicaller_frame
                del apicaller_module

            # Log only if the caller is not from the zhmcclient package
            return apicaller_module_name.split('.')[0] != 'zhmcclient'

        def blanked_dict(properties):
            """
            Return a copy of the properties dict, with blanked out values
            according to the 'blanked_properties' and 'properties_pos'
            arguments of the 'logged_api_call' decorator.
            """
            # properties may also be a DictView (subclass of Mapping)
            assert isinstance(properties, Mapping)
            copied_properties = dict(properties)
            for pn in blanked_properties:
                try:
                    copied_properties[pn] = BLANKED_OUT_STRING
                except KeyError:
                    pass
            return copied_properties

        def blanked_args(args, kwargs):
            """
            Return a copy of args and kwargs, whereby the 'properties' argument
            has items blanked out according to the 'blanked_properties' and
            'properties_pos' arguments of the 'logged_api_call' decorator.
            """
            logged_kwargs = dict(kwargs)
            logged_args = list(args)
            if blanked_properties is not None:
                if 'properties' in kwargs:
                    logged_kwargs['properties'] = \
                        blanked_dict(kwargs['properties'])
                else:
                    logged_args[properties_pos] = \
                        blanked_dict(args[properties_pos])
            return tuple(logged_args), logged_kwargs

        def log_call(args, kwargs):
            """
            Log the call to the API function.
            """
            logged_args, logged_kwargs = blanked_args(args, kwargs)
            logger.debug("Called: %s, args: %.500s, kwargs: %.500s",
                         apifunc_str,
                         log_escaped(repr(logged_args)),
                         log_escaped(repr(logged_kwargs)))

        def log_return(result):
            """
            Log the return from the API function.
            """
            logger.debug("Return: %s, result: %.1000s",
                         apifunc_str, log_escaped(repr(result)))

        @functools.wraps(func)
        def log_api_call(*args, **kwargs):
            """
            Log entry to and exit from the decorated function, at the debug
            level.

            Note that this wrapper function is called every time the decorated
            function/method is called, but that the log message only needs to
            be constructed when logging for this logger and for this log level
            is turned on. Therefore, we do as much as possible in the decorator
            function, plus we use %-formatting and lazy interpolation provided
            by the log functions, in order to save resources in this function
            here.

            Parameters:

              func (function object): The decorated function.

              *args: Any positional arguments for the decorated function.

              **kwargs: Any keyword arguments for the decorated function.
            """

            # Note that in this function, we are in the context where the
            # decorated function is actually called.
            _log_it = is_external_call() and logger.isEnabledFor(logging.DEBUG)

            if _log_it:
                log_call(args, kwargs)

            result = func(*args, **kwargs)  # The zhmcclient function

            if _log_it:
                log_return(result)

            return result

        return log_api_call

    # When the logged_api_call decorator is specified with its optional
    # arguments, org_func is None
    if org_func:
        return _decorate(org_func)
    return _decorate



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for _console module.
"""


import re
import pytest

from zhmcclient import Client, Error, Console, UserManager, UserRoleManager, \
    UserPatternManager, PasswordRuleManager, TaskManager, \
    LdapServerDefinitionManager
from zhmcclient_mock import FakedSession
from tests.common.utils import assert_resources


class TestConsole:
    """All tests for the Console and ConsoleManager classes."""

    def setup_method(self):
        """
        Setup that is called by pytest before each test method.

        Set up a faked session, and add a faked Console without any
        child resources.
        """
        # pylint: disable=attribute-defined-outside-init

        self.session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.client = Client(self.session)

        self.faked_console = self.session.hmc.consoles.add({
            'object-id': None,
            # object-uri will be automatically set
            'parent': None,
            'class': 'console',
            'name': 'fake-console1',
            'description': 'Console #1',
        })

    def test_consolemanager_initial_attrs(self):
        """Test initial attributes of ConsoleManager."""

        console_mgr = self.client.consoles

        # Verify all public properties of the manager object
        assert console_mgr.resource_class == Console
        assert console_mgr.class_name == 'console'
        assert console_mgr.session == self.session
        assert console_mgr.parent is None
        assert console_mgr.client == self.client

    # TODO: Test for ConsoleManager.__repr__()

    @pytest.mark.parametrize(
        "full_properties_kwargs, prop_names", [
            (dict(full_properties=False),
             ['object-uri']),
            (dict(full_properties=True),
             ['object-uri', 'name']),
            ({},  # test default for full_properties (False)
             ['object-uri']),
        ]
    )
    @pytest.mark.parametrize(
        "filter_args", [  # will be ignored
            None,
            {},
            {'name': 'foo'},
        ]
    )
    def test_consolemanager_list(
            self, filter_args, full_properties_kwargs, prop_names):
        """Test ConsoleManager.list()."""

        exp_faked_consoles = [self.faked_console]
        console_mgr = self.client.consoles

        # Execute the code to be tested
        consoles = console_mgr.list(
            filter_args=filter_args,
            **full_properties_kwargs)

        assert_resources(consoles, exp_faked_consoles, prop_names)

    def test_console_initial_attrs(self):
        """Test initial attributes of Console."""

        console_mgr = self.client.consoles
        console = console_mgr.find(name=self.faked_console.name)

        # Verify all public properties of the resource object (except those
        # of its BaseResource superclass which are tested at that level).
        assert isinstance(console.users, UserManager)
        assert isinstance(console.user_roles, UserRoleManager)
        assert isinstance(console.user_patterns, UserPatternManager)
        assert isinstance(console.password_rules, PasswordRuleManager)
        assert isinstance(console.tasks, TaskManager)
        assert isinstance(console.ldap_server_definitions,
                          LdapServerDefinitionManager)

    def test_console_repr(self):
        """Test Console.__repr__()."""

        console_mgr = self.client.consoles
        console = console_mgr.find(name=self.faked_console.name)

        # Execute the code to be tested
        repr_str = repr(console)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf'^{console.__class__.__name__}\s+at\s+'
            rf'0x{id(console):08x}\s+\(\\n.*',
            repr_str)

    @pytest.mark.parametrize(
        "wait", [
            True,
            # False,  # TODO: Re-enable once implemented
        ]
    )
    @pytest.mark.parametrize(
        "force", [
            True,
            False,
        ]
    )
    def test_console_restart(self, force, wait):
        """Test Console.restart()."""

        console_mgr = self.client.consoles
        console = console_mgr.find(name=self.faked_console.name)

        # Note: The force parameter is passed in, but we expect the restart
        # to always succeed. This means we are not testing cases with other
        # HMC users being logged on, that would either cause a non-forced
        # restart to be rejected, or a forced restart to succeed despite them.

        # Execute the code to be tested.
        ret = console.restart(
            force=force,
            wait_for_available=wait,
            operation_timeout=600)

        assert ret is None

        if wait:
            # The HMC is expected to be up again, and therefore a simple
            # operation is expected to succeed:
            try:
                self.client.query_api_version()
            except Error as exc:
                pytest.fail(
                    "Unexpected zhmcclient exception during "
                    f"query_api_version() after HMC restart: {exc}")
        else:
            # The HMC is expected to still be in the restart process, and
            # therefore a simple operation is expected to fail. This test
            # may end up to be timing dependent.
            try:
                self.client.query_api_version()
            except Error:
                pass
            except Exception as exc:  # pylint: disable=broad-except
                pytest.fail(
                    "Unexpected non-zhmcclient exception during "
                    f"query_api_version() after HMC restart: {exc}")
            else:
                pytest.fail(
                    "Unexpected success of query_api_version() after HMC "
                    "restart.")

    @pytest.mark.parametrize(
        "force", [
            True,
            False,
        ]
    )
    def test_console_shutdown(self, force):
        """Test Console.shutdown()."""

        console_mgr = self.client.consoles
        console = console_mgr.find(name=self.faked_console.name)

        # Note: The force parameter is passed in, but we expect the shutdown
        # to always succeed. This means we are not testing cases with other
        # HMC users being logged on, that would either cause a non-forced
        # shutdown to be rejected, or a forced shutdown to succeed despite
        # them.

        # Execute the code to be tested.
        ret = console.shutdown(force=force)

        assert ret is None

        # The HMC is expected to be offline, and therefore a simple operation
        # is expected to fail.
        try:
            self.client.query_api_version()
        except Error:
            pass
        except Exception as exc:  # pylint: disable=broad-except
            pytest.fail(
                "Unexpected non-zhmcclient exception during "
                f"query_api_version() after HMC shutdown: {exc}")
        else:
            pytest.fail(
                "Unexpected success of query_api_version() after HMC "
                "shutdown.")

    def test_console_audit_log(self):
        """Test Console.get_audit_log()."""

        # TODO: Add begin/end_time once mocked audit log is supported

        console_mgr = self.client.consoles
        console = console_mgr.find(name=self.faked_console.name)

        # Execute the code to be tested.
        log_items = console.get_audit_log()

        assert isinstance(log_items, list)

        # TODO: Verify log items once mocked audit log is supported

    def test_console_security_log(self):
        """Test Console.get_security_log()."""

        # TODO: Add begin/end_time once mocked security log is supported

        console_mgr = self.client.consoles
        console = console_mgr.find(name=self.faked_console.name)

        # Execute the code to be tested.
        log_items = console.get_security_log()

        assert isinstance(log_items, list)

        # TODO: Verify log items once mocked security log is supported

    @pytest.mark.parametrize(
        "name, exp_cpc_names, prop_names", [
            (None,
             ['ucpc_name_1', 'ucpc_name_2'],
             ['object-uri', 'name']),
            ('.*',
             ['ucpc_name_1', 'ucpc_name_2'],
             ['object-uri', 'name']),
            ('ucpc_name_.*',
             ['ucpc_name_1', 'ucpc_name_2'],
             ['object-uri', 'name']),
            ('ucpc_name_1',
             ['ucpc_name_1'],
             ['object-uri', 'name']),
            ('ucpc_name_1.*',
             ['ucpc_name_1'],
             ['object-uri', 'name']),
            ('ucpc_name_1.+',
             [],
             ['object-uri', 'name']),
        ]
    )
    def test_console_list_unmanaged_cpcs(
            self, name, exp_cpc_names, prop_names):
        """Test Console.list_unmanaged_cpcs() and
        UnmanagedCpcManager.list()."""

        console_mgr = self.client.consoles
        console = console_mgr.find(name=self.faked_console.name)

        # Add two unmanaged faked CPCs
        faked_ucpc1 = self.faked_console.unmanaged_cpcs.add({
            'object-id': 'ucpc-oid-1',
            # object-uri is set up automatically
            'name': 'ucpc_name_1',
        })
        faked_ucpc2 = self.faked_console.unmanaged_cpcs.add({
            'object-id': 'ucpc-oid-2',
            # object-uri is set up automatically
            'name': 'ucpc_name_2',
        })
        faked_ucpcs = [faked_ucpc1, faked_ucpc2]

        exp_faked_ucpcs = [cpc for cpc in faked_ucpcs
                           if cpc.name in exp_cpc_names]

        # Execute the code to be tested.
        # This indirectly tests UnmanagedCpcManager.list().
        ucpcs = console.list_unmanaged_cpcs(name)

        assert_resources(ucpcs, exp_faked_ucpcs, prop_names)



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for _notification module.

The test strategy is to mock the STOMP messages of the HMC using the
requests_mock package.
"""


import json
import threading
from collections import namedtuple
import logging
import time
from unittest.mock import patch
import pytest
import stomp

from zhmcclient._notification import NotificationReceiver
from zhmcclient._utils import stomp_uses_frames
from zhmcclient._exceptions import SubscriptionNotFound, \
    NotificationConnectionError
from zhmcclient._constants import JMS_LOGGER_NAME, \
    STOMP_MIN_CONNECTION_CHECK_TIME

# We use the JMS logger for logging. The log is reported by pytest in case
# of testcase errors when using pytest option:
# --log-level debug
# Because threads are involved, it makes sense to add the thread name in the
# log format using pytest option:
# --log-format "%(asctime)s %(threadName)s %(name)s %(levelname)s %(message)s
JMS_LOGGER = logging.getLogger(JMS_LOGGER_NAME)


def create_event_args(headers, message):
    """
    Transform headers, message to event method parameters.
    This is the inverse of get_headers_message().
    """
    if stomp_uses_frames(stomp.__version__):
        Frame = namedtuple('frame', ['headers', 'body'])
        frame = Frame(headers, message)
        frame_args = (frame,)
    else:
        frame_args = headers, message
    return frame_args


class MockedStompConnection:
    """
    A class that replaces stomp.Connection for the usage scope in the
    zhmcclient._notification module, and that adds the ability to
    queue STOMP messages.
    """

    def __init__(self, *args, **kwargs):
        # pylint: disable=unused-argument
        """
        For the mocked connection, we ignore the args (host/port, etc).
        """
        self._state_connected = False
        self._listener = None
        self._connect_userid = None
        self._connect_password = None
        self._connect_wait = None
        self._subscriptions = []  # items: tuple(dest, id, ack)
        self._queued_messages = []  # items: tuple(headers, message_str)
        self._sender_thread = None
        self._ssl_args = None
        self._ssl_kwargs = None

    def set_ssl(self, *args, **kwargs):
        # pylint: disable=unused-argument
        """
        Mocks the same-named method of stomp.Connection.
        """
        assert not self._state_connected
        self._ssl_args = args
        self._ssl_kwargs = kwargs

    def set_listener(self, name, listener):
        # pylint: disable=unused-argument
        """
        Mocks the same-named method of stomp.Connection.
        """
        assert not self._state_connected
        self._listener = listener

    def start(self):
        """
        Mocks the same-named method of stomp.Connection.
        """
        assert not self._state_connected

    def connect(self, userid, password, wait):
        """
        Mocks the same-named method of stomp.Connection.
        """
        JMS_LOGGER.debug(
            "test_notifcation: MockedStompConnection.connect() called")
        if not self._state_connected:
            self._state_connected = True
            self._connect_userid = userid
            self._connect_password = password
            self._connect_wait = wait

    def is_connected(self):
        """
        Mocks the same-named method of stomp.Connection.
        """
        return self._state_connected

    def subscribe(self, destination, id, ack):
        # pylint: disable=redefined-builtin
        """
        Mocks the same-named method of stomp.Connection.
        """
        assert self._state_connected
        self._subscriptions.append((destination, id, ack))

    def unsubscribe(self, id):
        # pylint: disable=redefined-builtin
        """
        Mocks the same-named method of stomp.Connection.
        """
        assert self._state_connected
        for _dest, _id, _ack in self._subscriptions:
            if _id == id:
                self._subscriptions.remove((_dest, _id, _ack))

    def disconnect(self, receipt=None):
        # pylint: disable=unused-argument
        """
        Mocks the same-named method of stomp.Connection.
        """
        JMS_LOGGER.debug(
            "test_notifcation: MockedStompConnection.disconnect() called")
        if self._state_connected:
            self._state_connected = False

    def mock_add_message(self, headers, message):
        """
        Adds a STOMP message to the queue.
        """
        assert self._sender_thread is None
        if not isinstance(message, str):
            message = json.dumps(message)
        self._queued_messages.append((headers, message))

    def mock_start(self):
        """
        Start the STOMP message sender thread.

        This can be done independent of the connection state.
        """
        assert self._sender_thread is None
        self._sender_thread = threading.Thread(target=self.mock_sender_run)
        JMS_LOGGER.debug("test_notifcation: Starting mock sender thread")
        self._sender_thread.start()

    def mock_stop(self):
        """
        Wait for the STOMP message sender thread to finish.

        This can be done independent of the connection state.
        """
        assert self._sender_thread is not None
        JMS_LOGGER.debug(
            "test_notifcation: Waiting for mock sender thread to finish")
        self._sender_thread.join(STOMP_MIN_CONNECTION_CHECK_TIME + 2)
        self._sender_thread = None
        JMS_LOGGER.debug(
            "test_notifcation: Mock sender thread finished")

    def mock_sender_run(self):
        """
        Thread function that runs in the mock sender thread.

        It simulates the HMC sending STOMP messages by sending the previously
        queued STOMP messages to the notification listener.

        It returns when all messages have been sent and thus ends the mock
        sender thread in which it runs.
        """
        JMS_LOGGER.debug(
            "test_notifcation: Mock sender thread gets control")
        for msg_item in self._queued_messages:
            # The following method blocks until it can deliver a message
            headers, message_str = msg_item
            frame_args = create_event_args(headers, message_str)
            self._listener.on_message(*frame_args)
        self._listener.on_disconnected()
        JMS_LOGGER.debug(
            "test_notifcation: Mock sender thread is done")

    def mock_get_subscription(self, topic):
        """
        Find the subscription with the specified topic name and return it.
        """
        for _dest, _id, _ack in self._subscriptions:
            dest = '/topic/' + topic
            if _dest == dest:
                return (_dest, _id, _ack)
        return None


def receiver_run(receiver, msg_items):
    """
    Thread function that will be run in the receiver thread.

    It invokes receiver.notifications() to get notifications which are
    appended to the msg_items list.

    It ends when the main thread calls receiver.close() which causes
    receiver.notifications() to return and thus end the loop.
    """
    JMS_LOGGER.debug(
        "test_notifcation: Receiver thread gets control")
    try:
        for headers, message in receiver.notifications():
            msg_items.append((headers, message))
    except NotificationConnectionError:
        # This happens at the end of some testcases.
        pass
    JMS_LOGGER.debug(
        "test_notifcation: Receiver thread is done")


def receive_notifications(receiver):
    """
    Start a thread running the receiver function and wait for its completion.
    """
    msg_items = []
    receiver_thread = threading.Thread(target=receiver_run,
                                       args=(receiver, msg_items))
    JMS_LOGGER.debug(
        "test_notifcation: Starting receiver thread")
    receiver_thread.start()
    # Allow for some time to process the received messages
    time.sleep(0.5)
    JMS_LOGGER.debug(
        "test_notifcation: Triggering receiver close")
    receiver.close()
    JMS_LOGGER.debug(
        "test_notifcation: Waiting for receiver thread to finish")
    receiver_thread.join(STOMP_MIN_CONNECTION_CHECK_TIME + 1)
    if receiver_thread.is_alive():
        raise AssertionError("receiver_thread is still alive")
    return msg_items


class TestNotificationOneTopic:
    """
    Test class for one notification topic.
    """

    def setup_method(self):
        """
        Setup that is called by pytest before each test method.
        """
        # pylint: disable=attribute-defined-outside-init

        self.topic = 'fake-topic'
        self.hmc = 'fake-hmc'
        self.userid = 'fake-userid'
        self.password = 'fake-password'
        self.std_headers = {
            'notification-type': 'fake-type'
        }

    def setup_receiver(self):
        """Set up the notification receiver for the test"""
        receiver = NotificationReceiver(self.topic, self.hmc, self.userid,
                                        self.password)
        receiver.connect()
        mocked_conn = receiver._conn  # pylint: disable=protected-access
        assert isinstance(mocked_conn, MockedStompConnection)
        return receiver, mocked_conn

    @patch(target='stomp.Connection', new=MockedStompConnection)
    def test_no_messages(self):
        """Test function for not receiving any notification."""

        receiver, mocked_conn = self.setup_receiver()

        # We do not add any STOMP messages

        mocked_conn.mock_start()  # pylint: disable=no-member
        msg_items = receive_notifications(receiver)
        mocked_conn.mock_stop()  # pylint: disable=no-member

        assert msg_items == []

    @patch(target='stomp.Connection', new=MockedStompConnection)
    def test_one_message(self):
        """Test function for receiving one notification."""

        receiver, mocked_conn = self.setup_receiver()

        # Add one STOMP message to be sent
        message_obj = dict(a=1, b=2)
        # pylint: disable=no-member
        mocked_conn.mock_add_message(self.std_headers, message_obj)

        mocked_conn.mock_start()  # pylint: disable=no-member
        msg_items = receive_notifications(receiver)
        mocked_conn.mock_stop()  # pylint: disable=no-member

        assert len(msg_items) == 1

        msg0 = msg_items[0]
        assert msg0[0] == self.std_headers
        assert msg0[1] == message_obj


class TestNotificationTwoTopics:
    """
    Test class for two notification topics.
    """

    def setup_method(self):
        """
        Setup that is called by pytest before each test method.
        """
        # pylint: disable=attribute-defined-outside-init

        self.topics = ('fake-topic1', 'fake-topic2')
        self.hmc = 'fake-hmc'
        self.userid = 'fake-userid'
        self.password = 'fake-password'
        self.std_headers = {
            'notification-type': 'fake-type'
        }

    def setup_receiver(self):
        """Set up the notification receiver for the test"""
        receiver = NotificationReceiver(self.topics, self.hmc, self.userid,
                                        self.password)
        receiver.connect()
        mocked_conn = receiver._conn  # pylint: disable=protected-access
        assert isinstance(mocked_conn, MockedStompConnection)
        return receiver, mocked_conn

    @patch(target='stomp.Connection', new=MockedStompConnection)
    def test_no_messages(self):
        """Test function for not receiving any notification."""

        receiver, mocked_conn = self.setup_receiver()

        # We do not add any STOMP messages

        mocked_conn.mock_start()  # pylint: disable=no-member
        msg_items = receive_notifications(receiver)
        mocked_conn.mock_stop()  # pylint: disable=no-member

        assert msg_items == []

    @patch(target='stomp.Connection', new=MockedStompConnection)
    def test_one_message(self):
        """Test function for receiving one notification."""

        receiver, mocked_conn = self.setup_receiver()

        # Add one STOMP message to be sent
        message_obj = dict(a=1, b=2)
        # pylint: disable=no-member
        mocked_conn.mock_add_message(self.std_headers, message_obj)

        mocked_conn.mock_start()  # pylint: disable=no-member
        msg_items = receive_notifications(receiver)
        mocked_conn.mock_stop()  # pylint: disable=no-member

        assert len(msg_items) == 1

        msg0 = msg_items[0]
        assert msg0[0] == self.std_headers
        assert msg0[1] == message_obj


class TestNotificationSubscriptionMgmt:
    """
    Test class for subscription management.
    """

    def setup_method(self):
        """
        Setup that is called by pytest before each test method.
        """
        # pylint: disable=attribute-defined-outside-init

        self.topics = ('fake-topic1', 'fake-topic2')
        self.hmc = 'fake-hmc'
        self.userid = 'fake-userid'
        self.password = 'fake-password'
        self.std_headers = {
            'notification-type': 'fake-type'
        }

    def setup_receiver(self):
        """Set up the notification receiver for the test"""
        receiver = NotificationReceiver(self.topics, self.hmc, self.userid,
                                        self.password)
        receiver.connect()
        mocked_conn = receiver._conn  # pylint: disable=protected-access
        assert isinstance(mocked_conn, MockedStompConnection)
        return receiver, mocked_conn

    @patch(target='stomp.Connection', new=MockedStompConnection)
    def test_is_subscribed(self):
        """Test function for is_subscribed() method."""

        receiver, mocked_conn = self.setup_receiver()

        # pylint: disable=no-member
        assert mocked_conn.mock_get_subscription('fake-topic1')
        # pylint: disable=no-member
        assert mocked_conn.mock_get_subscription('fake-topic2')

        result = receiver.is_subscribed('fake-topic1')
        assert result is True

        result = receiver.is_subscribed('foo')
        assert result is False

    @patch(target='stomp.Connection', new=MockedStompConnection)
    def test_get_subscription(self):
        """Test function for get_subscription() method."""

        receiver, mocked_conn = self.setup_receiver()

        id_value1 = receiver.get_subscription('fake-topic1')
        id_value2 = receiver.get_subscription('fake-topic2')
        assert id_value1 != id_value2

        # Check that the subscriptions are still in place
        # pylint: disable=no-member
        assert mocked_conn.mock_get_subscription('fake-topic1')
        # pylint: disable=no-member
        assert mocked_conn.mock_get_subscription('fake-topic2')

    @patch(target='stomp.Connection', new=MockedStompConnection)
    def test_get_subscription_nonexisting(self):
        """Test function for get_subscription() method for a non-existing
        subscription.."""

        receiver, mocked_conn = self.setup_receiver()

        with pytest.raises(SubscriptionNotFound):
            receiver.get_subscription('bla')

        # Check that the subscriptions are still in place
        # pylint: disable=no-member
        assert mocked_conn.mock_get_subscription('fake-topic1')
        # pylint: disable=no-member
        assert mocked_conn.mock_get_subscription('fake-topic2')

    @patch(target='stomp.Connection', new=MockedStompConnection)
    def test_subscribe(self):
        """Test function for subscribe() method."""

        receiver, mocked_conn = self.setup_receiver()

        receiver.subscribe('foo')

        # pylint: disable=no-member
        assert mocked_conn.mock_get_subscription('foo')

        # Check that the subscriptions are still in place
        # pylint: disable=no-member
        assert mocked_conn.mock_get_subscription('fake-topic1')
        # pylint: disable=no-member
        assert mocked_conn.mock_get_subscription('fake-topic2')

    @patch(target='stomp.Connection', new=MockedStompConnection)
    def test_unsubscribe(self):
        """Test function for unsubscribe() method."""

        receiver, mocked_conn = self.setup_receiver()

        receiver.unsubscribe('fake-topic1')

        # pylint: disable=no-member
        assert not mocked_conn.mock_get_subscription('fake-topic1')

        # Check that the subscriptions are still in place
        # pylint: disable=no-member
        assert mocked_conn.mock_get_subscription('fake-topic2')

    @patch(target='stomp.Connection', new=MockedStompConnection)
    def test_unsubscribe_nonexisting(self):
        """Test function for unsubscribe() for a non-existing subscription."""

        receiver, mocked_conn = self.setup_receiver()

        with pytest.raises(SubscriptionNotFound):
            receiver.unsubscribe('bla')

        # Check that the subscriptions are still in place
        # pylint: disable=no-member
        assert mocked_conn.mock_get_subscription('fake-topic1')
        # pylint: disable=no-member
        assert mocked_conn.mock_get_subscription('fake-topic2')



# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for _capacity_group module.
"""


import re
import copy
import pytest

from zhmcclient import Client, CapacityGroup, CapacityGroupManager, \
    HTTPError, NotFound
from zhmcclient_mock import FakedSession
from tests.common.utils import assert_resources


# Object IDs and names of our faked capacity groups:
CPC_OID = 'fake-cpc1-oid'
CPC_URI = f'/api/cpcs/{CPC_OID}'
CG1_OID = 'cg1-oid'
CG1_NAME = 'cg 1'
CG1_SHORT_NAME = 'CG1'
CG2_OID = 'cg2-oid'
CG2_NAME = 'cg 2'
CG2_SHORT_NAME = 'CG2'


class TestCapacityGroup:
    """All tests for the CapacityGroup and CapacityGroupManager classes."""

    def setup_method(self):
        """
        Setup that is called by pytest before each test method.

        Set up a faked session, and add a faked CPC in DPM mode without any
        child resources.
        """
        # pylint: disable=attribute-defined-outside-init

        self.session = FakedSession('fake-host', 'fake-hmc', '2.14.1', '1.9')
        self.client = Client(self.session)
        self.faked_cpc = self.session.hmc.cpcs.add({
            'object-id': CPC_OID,
            # object-uri is set up automatically
            'parent': None,
            'class': 'cpc',
            'name': 'fake-cpc1-name',
            'description': 'CPC #1 (DPM mode)',
            'status': 'active',
            'dpm-enabled': True,
            'is-ensemble-member': False,
            'iml-mode': 'dpm',
        })
        assert self.faked_cpc.uri == CPC_URI
        self.cpc = self.client.cpcs.find(name='fake-cpc1-name')
        self.faked_console = self.session.hmc.consoles.add({
            # object-id is set up automatically
            # object-uri is set up automatically
            # parent will be automatically set
            # class will be automatically set
            'name': 'fake-console-name',
            'description': 'The HMC',
        })
        self.console = self.client.consoles.console

    def add_capacity_group1(self):
        """Add capacity group 1"""

        faked_capacity_group = self.faked_cpc.capacity_groups.add({
            'element-id': CG1_OID,
            # element-uri will be automatically set
            # parent will be automatically set
            # class will be automatically set
            'name': CG1_NAME,
            'short-name': CG1_SHORT_NAME,
            'description': 'Capacity Group #1',
            'capping-enabled': True,
            'absolute-general-purpose-proc-cap': 0.0,
            'absolute-ifl-proc-cap': 100.0,
        })
        return faked_capacity_group

    def add_capacity_group2(self):
        """Add capacity group 2"""

        faked_capacity_group = self.faked_cpc.capacity_groups.add({
            'element-id': CG2_OID,
            # element-uri will be automatically set
            # parent will be automatically set
            # class will be automatically set
            'name': CG2_NAME,
            'short-name': CG2_SHORT_NAME,
            'description': 'Capacity Group #2',
            'capping-enabled': True,
            'absolute-general-purpose-proc-cap': 0.0,
            'absolute-ifl-proc-cap': 100.0,
        })
        return faked_capacity_group

    def add_partition1(self):
        """Add partition 1 (shared processor mode)"""

        faked_partition = self.faked_cpc.partitions.add({
            'object-id': 'fake-part1-oid',
            # object-uri will be automatically set
            'parent': self.faked_cpc.uri,
            'class': 'partition',
            'name': 'fake-part1-name',
            'description': 'Partition #1 (shared)',
            'status': 'active',
            'processor-mode': 'shared',
            'initial-memory': 1024,
            'maximum-memory': 2048,
        })
        return faked_partition

    def add_partition2(self):
        """Add partition 2 (dedicated processor mode)"""

        faked_partition = self.faked_cpc.partitions.add({
            'object-id': 'fake-part2-oid',
            # object-uri will be automatically set
            'parent': self.faked_cpc.uri,
            'class': 'partition',
            'name': 'fake-part2-name',
            'description': 'Partition #2 (dedicated)',
            'status': 'active',
            'processor-mode': 'dedicated',
            'initial-memory': 1024,
            'maximum-memory': 2048,
        })
        return faked_partition

    def test_cgm_initial_attrs(self):
        """Test initial attributes of CapacityGroupManager."""

        capacity_group_mgr = self.cpc.capacity_groups

        assert isinstance(capacity_group_mgr, CapacityGroupManager)

        # Verify all public properties of the manager object
        assert capacity_group_mgr.resource_class == CapacityGroup
        assert capacity_group_mgr.session == self.session
        assert capacity_group_mgr.parent == self.cpc

    # TODO: Test for CapacityGroupManager.__repr__()

    testcases_cgm_list_full_properties = (
        "full_properties_kwargs, prop_names", [
            ({},
             ['element-uri', 'name']),
            (dict(full_properties=False),
             ['element-uri', 'name']),
            (dict(full_properties=True),
             None),
        ]
    )

    @pytest.mark.parametrize(
        *testcases_cgm_list_full_properties
    )
    def test_cgm_list_full_properties(
            self, full_properties_kwargs, prop_names):
        """Test CapacityGroupManager.list() with full_properties."""

        # Add two faked capacity groups
        faked_capacity_group1 = self.add_capacity_group1()
        faked_capacity_group2 = self.add_capacity_group2()

        exp_faked_capacity_groups = [faked_capacity_group1,
                                     faked_capacity_group2]
        capacity_group_mgr = self.cpc.capacity_groups

        # Execute the code to be tested
        capacity_groups = capacity_group_mgr.list(**full_properties_kwargs)

        assert_resources(capacity_groups, exp_faked_capacity_groups, prop_names)

    testcases_cgm_list_filter_args = (
        "filter_args, exp_names", [
            ({'element-id': CG1_OID},
             [CG1_NAME]),
            ({'element-id': CG2_OID},
             [CG2_NAME]),
            ({'element-id': [CG1_OID, CG2_OID]},
             [CG1_NAME, CG2_NAME]),
            ({'element-id': [CG1_OID, CG1_OID]},
             [CG1_NAME]),
            ({'element-id': CG1_OID + 'foo'},
             []),
            ({'element-id': [CG1_OID, CG2_OID + 'foo']},
             [CG1_NAME]),
            ({'element-id': [CG2_OID + 'foo', CG1_OID]},
             [CG1_NAME]),
            ({'name': CG1_NAME},
             [CG1_NAME]),
            ({'name': CG2_NAME},
             [CG2_NAME]),
            ({'name': [CG1_NAME, CG2_NAME]},
             [CG1_NAME, CG2_NAME]),
            ({'name': CG1_NAME + 'foo'},
             []),
            ({'name': [CG1_NAME, CG2_NAME + 'foo']},
             [CG1_NAME]),
            ({'name': [CG2_NAME + 'foo', CG1_NAME]},
             [CG1_NAME]),
            ({'name': [CG1_NAME, CG1_NAME]},
             [CG1_NAME]),
            ({'name': '.*cg 1'},
             [CG1_NAME]),
            ({'name': 'cg 1.*'},
             [CG1_NAME]),
            ({'name': 'cg .'},
             [CG1_NAME, CG2_NAME]),
            ({'name': '.g 1'},
             [CG1_NAME]),
            ({'name': '.+'},
             [CG1_NAME, CG2_NAME]),
            ({'name': 'cg 1.+'},
             []),
            ({'name': '.+cg 1'},
             []),
            ({'name': CG1_NAME,
              'element-id': CG1_OID},
             [CG1_NAME]),
            ({'name': CG1_NAME,
              'element-id': CG1_OID + 'foo'},
             []),
            ({'name': CG1_NAME + 'foo',
              'element-id': CG1_OID},
             []),
            ({'name': CG1_NAME + 'foo',
              'element-id': CG1_OID + 'foo'},
             []),
        ]
    )

    @pytest.mark.parametrize(
        *testcases_cgm_list_filter_args
    )
    def test_cgm_list_filter_args(
            self, filter_args, exp_names):
        """Test CapacityGroupManager.list() with filter_args."""

        # Add two faked capacity_groups
        self.add_capacity_group1()
        self.add_capacity_group2()

        capacity_group_mgr = self.cpc.capacity_groups

        # Execute the code to be tested
        capacity_groups = capacity_group_mgr.list(filter_args=filter_args)

        assert len(capacity_groups) == len(exp_names)
        if exp_names:
            names = [p.properties['name'] for p in capacity_groups]
            assert set(names) == set(exp_names)

    testcases_cgm_create = (
        "input_props, exp_prop_names, exp_exc", [
            ({'name': 'fake-cg-x',
              'capping-enabled': True,
              'absolute-ifl-proc-cap': 50.0},
             ['element-uri', 'name', 'capping-enabled',
              'absolute-ifl-proc-cap'],
             None),
        ]
    )

    @pytest.mark.parametrize(
        *testcases_cgm_create
    )
    def test_cgm_create(
            self, input_props, exp_prop_names, exp_exc):
        """Test CapacityGroupManager.create()."""

        capacity_group_mgr = self.cpc.capacity_groups

        if exp_exc is not None:

            with pytest.raises(exp_exc.__class__) as exc_info:

                # Execute the code to be tested
                capacity_group = capacity_group_mgr.create(
                    properties=input_props)

            exc = exc_info.value
            if isinstance(exp_exc, HTTPError):
                assert exc.http_status == exp_exc.http_status
                assert exc.reason == exp_exc.reason

        else:

            # Execute the code to be tested.
            # Note: the CapacityGroup object returned by CapacityGroup.create()
            # has the input properties plus 'element-uri'.
            capacity_group = capacity_group_mgr.create(properties=input_props)

            # Check the resource for consistency within itself
            assert isinstance(capacity_group, CapacityGroup)
            capacity_group_name = capacity_group.name
            exp_capacity_group_name = capacity_group.properties['name']
            assert capacity_group_name == exp_capacity_group_name
            capacity_group_uri = capacity_group.uri
            exp_capacity_group_uri = capacity_group.properties['element-uri']
            assert capacity_group_uri == exp_capacity_group_uri

            # Check the properties against the expected names and values
            for prop_name in exp_prop_names:
                assert prop_name in capacity_group.properties
                if prop_name in input_props:
                    value = capacity_group.properties[prop_name]
                    exp_value = input_props[prop_name]
                    assert value == exp_value

    def test_cg_repr(self):
        """Test CapacityGroup.__repr__()."""

        # Add a faked capacity_group
        faked_capacity_group = self.add_capacity_group1()

        capacity_group_mgr = self.cpc.capacity_groups
        capacity_group = capacity_group_mgr.find(name=faked_capacity_group.name)

        # Execute the code to be tested
        repr_str = repr(capacity_group)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf'^{capacity_group.__class__.__name__}\s+at\s+'
            rf'0x{id(capacity_group):08x}\s+\(\\n.*',
            repr_str)

    def test_cg_delete_empty(self):
        """Test CapacityGroup.delete() of empty capacity group."""

        # Add a faked capacity group to be tested and another one
        faked_capacity_group = self.add_capacity_group1()
        self.add_capacity_group2()

        capacity_group_mgr = self.cpc.capacity_groups

        capacity_group = capacity_group_mgr.find(name=faked_capacity_group.name)

        # Execute the code to be tested.
        capacity_group.delete()

        # Check that the capacity group no longer exists
        with pytest.raises(NotFound):
            capacity_group_mgr.find(name=faked_capacity_group.name)

    def test_cg_delete_non_empty(self):
        """Test CapacityGroup.delete() of non-empty capacity group."""

        # Add a faked capacity group to be tested and another one
        faked_capacity_group = self.add_capacity_group1()
        self.add_capacity_group2()

        # Add a faked partition
        partition1 = self.add_partition1()

        capacity_group_mgr = self.cpc.capacity_groups

        capacity_group = capacity_group_mgr.find(name=faked_capacity_group.name)

        # Add the partition to the capacity group
        capacity_group.add_partition(partition1)

        with pytest.raises(HTTPError):

            # Execute the code to be tested.
            capacity_group.delete()

    def test_cg_delete_create_same(self):
        """Test CapacityGroup.delete() followed by create() with same name."""

        # Add a faked capacity_group to be tested and another one
        faked_capacity_group = self.add_capacity_group1()
        capacity_group_name = faked_capacity_group.name
        self.add_capacity_group2()

        # Construct the input properties for a third capacity_group
        cg3_props = copy.deepcopy(faked_capacity_group.properties)
        cg3_props['description'] = 'Third capacity_group'

        capacity_group_mgr = self.cpc.capacity_groups
        capacity_group = capacity_group_mgr.find(name=capacity_group_name)

        # Execute the deletion code to be tested.
        capacity_group.delete()

        # Check that the capacity_group no longer exists
        with pytest.raises(NotFound):
            capacity_group_mgr.find(name=capacity_group_name)

        # Execute the creation code to be tested.
        capacity_group_mgr.create(cg3_props)

        # Check that the capacity_group exists again under that name
        capacity_group3 = capacity_group_mgr.find(name=capacity_group_name)
        description = capacity_group3.get_property('description')
        assert description == 'Third capacity_group'

    testcases_cg_update_properties_names = (
        "capacity_group_name", [
            CG1_NAME,
            CG2_NAME,
        ]
    )

    testcases_cg_update_properties_props = (
        "input_props", [
            {},
            {'description': 'New capacity_group description'},
            {'shared': True},
            {'connectivity': 8},
        ]
    )

    @pytest.mark.parametrize(
        *testcases_cg_update_properties_names
    )
    @pytest.mark.parametrize(
        *testcases_cg_update_properties_props
    )
    def test_cg_update_properties(
            self, input_props, capacity_group_name):
        """Test CapacityGroup.update_properties()."""

        # Add faked capacity_groups
        self.add_capacity_group1()
        self.add_capacity_group2()

        capacity_group_mgr = self.cpc.capacity_groups
        capacity_group = capacity_group_mgr.find(name=capacity_group_name)

        capacity_group.pull_full_properties()
        saved_properties = copy.deepcopy(capacity_group.properties)

        # Execute the code to be tested
        capacity_group.update_properties(properties=input_props)

        # Verify that the resource object already reflects the property
        # updates.
        for prop_name in saved_properties:
            if prop_name in input_props:
                exp_prop_value = input_props[prop_name]
            else:
                exp_prop_value = saved_properties[prop_name]
            assert prop_name in capacity_group.properties
            prop_value = capacity_group.properties[prop_name]
            assert prop_value == exp_prop_value

        # Refresh the resource object and verify that the resource object
        # still reflects the property updates.
        capacity_group.pull_full_properties()
        for prop_name in saved_properties:
            if prop_name in input_props:
                exp_prop_value = input_props[prop_name]
            else:
                exp_prop_value = saved_properties[prop_name]
            assert prop_name in capacity_group.properties
            prop_value = capacity_group.properties[prop_name]
            assert prop_value == exp_prop_value

    def test_cg_update_name(self):
        """
        Test CapacityGroup.update_properties() with 'name' property.
        """

        # Add a faked capacity_group
        faked_capacity_group = self.add_capacity_group1()
        capacity_group_name = faked_capacity_group.name

        capacity_group_mgr = self.cpc.capacity_groups
        capacity_group = capacity_group_mgr.find(name=capacity_group_name)

        new_capacity_group_name = "new-" + capacity_group_name

        # Execute the code to be tested
        capacity_group.update_properties(
            properties={'name': new_capacity_group_name})

        # Verify that the resource is no longer found by its old name, using
        # list() (this does not use the name-to-URI cache).
        capacity_groups_list = capacity_group_mgr.list(
            filter_args=dict(name=capacity_group_name))
        assert len(capacity_groups_list) == 0

        # Verify that the resource is no longer found by its old name, using
        # find() (this uses the name-to-URI cache).
        with pytest.raises(NotFound):
            capacity_group_mgr.find(name=capacity_group_name)

        # Verify that the resource object already reflects the update, even
        # though it has not been refreshed yet.
        assert capacity_group.properties['name'] == new_capacity_group_name

        # Refresh the resource object and verify that it still reflects the
        # update.
        capacity_group.pull_full_properties()
        assert capacity_group.properties['name'] == new_capacity_group_name

        # Verify that the resource can be found by its new name, using find()
        new_capacity_group_find = capacity_group_mgr.find(
            name=new_capacity_group_name)
        assert new_capacity_group_find.properties['name'] == \
            new_capacity_group_name

        # Verify that the resource can be found by its new name, using list()
        new_capacity_groups_list = capacity_group_mgr.list(
            filter_args=dict(name=new_capacity_group_name))
        assert len(new_capacity_groups_list) == 1
        new_capacity_group_list = new_capacity_groups_list[0]
        assert new_capacity_group_list.properties['name'] == \
            new_capacity_group_name

    def test_cg_add_partition_empty(self):
        """Test CapacityGroup.add_partition() into empty capacity group."""

        # Add faked capacity_groups
        faked_cg1 = self.add_capacity_group1()
        self.add_capacity_group2()

        # Add faked partitions
        faked_part1 = self.add_partition1()

        cg1 = self.cpc.capacity_groups.find(name=faked_cg1.name)
        part1 = self.cpc.partitions.find(name=faked_part1.name)

        # Execute the code to be tested
        cg1.add_partition(part1)

        cg1.pull_full_properties()
        assert cg1.properties['partition-uris'] == [part1.uri]

    def test_cg_add_partition_same(self):
        """Test CapacityGroup.add_partition() into capacity group that
        contains the partition already."""

        # Add faked capacity_groups
        faked_cg1 = self.add_capacity_group1()
        self.add_capacity_group2()

        # Add faked partitions
        faked_part1 = self.add_partition1()

        cg1 = self.cpc.capacity_groups.find(name=faked_cg1.name)
        part1 = self.cpc.partitions.find(name=faked_part1.name)
        cg1.add_partition(part1)

        with pytest.raises(HTTPError) as exc_info:

            # Execute the code to be tested
            cg1.add_partition(part1)

        exc = exc_info.value
        assert exc.http_status == 409
        assert exc.reason == 130

    def test_cg_add_partition_other(self):
        """Test CapacityGroup.add_partition() with partition that is already
        in another capacity group."""

        # Add faked capacity_groups
        faked_cg1 = self.add_capacity_group1()
        faked_cg2 = self.add_capacity_group2()

        # Add faked partitions
        faked_part1 = self.add_partition1()

        cg1 = self.cpc.capacity_groups.find(name=faked_cg1.name)
        cg2 = self.cpc.capacity_groups.find(name=faked_cg2.name)
        part1 = self.cpc.partitions.find(name=faked_part1.name)
        cg2.add_partition(part1)

        with pytest.raises(HTTPError) as exc_info:

            # Execute the code to be tested
            cg1.add_partition(part1)

        exc = exc_info.value
        assert exc.http_status == 409
        assert exc.reason == 120

    def test_cg_add_partition_dedicated(self):
        """Test CapacityGroup.add_partition() with partition that has dedicated
        processor mode."""

        # Add faked capacity_groups
        faked_cg1 = self.add_capacity_group1()
        self.add_capacity_group2()

        # Add faked partitions
        faked_part2 = self.add_partition2()

        cg1 = self.cpc.capacity_groups.find(name=faked_cg1.name)
        part2 = self.cpc.partitions.find(name=faked_part2.name)

        with pytest.raises(HTTPError) as exc_info:

            # Execute the code to be tested
            cg1.add_partition(part2)

        exc = exc_info.value
        assert exc.http_status == 409
        assert exc.reason == 170

    def test_cg_remove_partition_success(self):
        """Test CapacityGroup.remove_partition(), simple success case."""

        # Add faked capacity_groups
        faked_cg1 = self.add_capacity_group1()
        self.add_capacity_group2()

        # Add faked partitions
        faked_part1 = self.add_partition1()

        cg1 = self.cpc.capacity_groups.find(name=faked_cg1.name)
        part1 = self.cpc.partitions.find(name=faked_part1.name)
        cg1.add_partition(part1)

        # Execute the code to be tested
        cg1.remove_partition(part1)

        cg1.pull_full_properties()
        assert cg1.properties['partition-uris'] == []

    def test_cg_remove_partition_empty(self):
        """Test CapacityGroup.remove_partition(), existing partition removal
        from empty capacity group."""

        # Add faked capacity_groups
        faked_cg1 = self.add_capacity_group1()
        self.add_capacity_group2()

        # Add faked partitions
        faked_part1 = self.add_partition1()

        cg1 = self.cpc.capacity_groups.find(name=faked_cg1.name)
        part1 = self.cpc.partitions.find(name=faked_part1.name)

        with pytest.raises(HTTPError) as exc_info:

            # Execute the code to be tested
            cg1.remove_partition(part1)

        exc = exc_info.value
        assert exc.http_status == 409
        assert exc.reason == 140



# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for user patterns (on CPCs in DPM mode).

These tests do not change any existing user patterns, but create,
modify and delete test user patterns.
"""


import warnings
import pytest
from requests.packages import urllib3

import zhmcclient
# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import

from .utils import skip_warn, pick_test_resources, TEST_PREFIX, \
    runtest_find_list, runtest_get_properties

urllib3.disable_warnings()

# Properties in minimalistic UserPattern objects (e.g. find_by_name())
UPATT_MINIMAL_PROPS = ['element-uri', 'name']

# Properties in UserPattern objects returned by list() without full props
UPATT_LIST_PROPS = ['element-uri', 'name', 'type']

# Properties whose values can change between retrievals of UserPattern objects
UPATT_VOLATILE_PROPS = []


def test_upatt_find_list(hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test list(), find(), findall().
    """
    client = zhmcclient.Client(hmc_session)
    console = client.consoles.console
    hd = hmc_session.hmc_definition

    api_version = client.query_api_version()
    hmc_version = api_version['hmc-version']
    hmc_version_info = tuple(map(int, hmc_version.split('.')))
    if hmc_version_info < (2, 13, 0):
        skip_warn(
            f"HMC {hd.host} of version {hmc_version} does not yet support "
            "user patterns")

    # Pick the user patterns to test with
    upatt_list = console.user_patterns.list()
    if not upatt_list:
        skip_warn(f"No user patterns defined on HMC {hd.host}")
    upatt_list = pick_test_resources(upatt_list)

    for upatt in upatt_list:
        print(f"Testing with user pattern {upatt.name!r}")
        runtest_find_list(
            hmc_session, console.user_patterns, upatt.name, 'name',
            'element-uri', UPATT_VOLATILE_PROPS, UPATT_MINIMAL_PROPS,
            UPATT_LIST_PROPS)


def test_upatt_property(hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test property related methods
    """
    client = zhmcclient.Client(hmc_session)
    console = client.consoles.console
    hd = hmc_session.hmc_definition

    api_version = client.query_api_version()
    hmc_version = api_version['hmc-version']
    hmc_version_info = tuple(map(int, hmc_version.split('.')))
    if hmc_version_info < (2, 13, 0):
        skip_warn(
            f"HMC {hd.host} of version {hmc_version} does not yet support "
            "user patterns")

    # Pick the user patterns to test with
    upatt_list = console.user_patterns.list()
    if not upatt_list:
        skip_warn(f"No user patterns defined on HMC {hd.host}")
    upatt_list = pick_test_resources(upatt_list)

    for upatt in upatt_list:
        print(f"Testing with user pattern {upatt.name!r}")

        # Select a property that is not returned by list()
        non_list_prop = 'description'

        runtest_get_properties(upatt.manager, non_list_prop)


def test_upatt_crud(hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test create, read, update and delete a user pattern.
    """
    client = zhmcclient.Client(hmc_session)
    console = client.consoles.console
    hd = hmc_session.hmc_definition

    api_version = client.query_api_version()
    hmc_version = api_version['hmc-version']
    hmc_version_info = tuple(map(int, hmc_version.split('.')))
    if hmc_version_info < (2, 13, 0):
        skip_warn(
            f"HMC {hd.host} of version {hmc_version} does not yet support "
            "user patterns")

    upatt_name = TEST_PREFIX + ' test_upatt_crud upatt1'
    upatt_name_new = upatt_name + ' new'

    # Ensure a clean starting point for this test
    try:
        upatt = console.user_patterns.find(name=upatt_name)
    except zhmcclient.NotFound:
        pass
    else:
        warnings.warn(
            f"Deleting test user pattern from previous run: {upatt_name!r}",
            UserWarning)
        upatt.delete()

    # Pick a template user to be the template user for the user pattern
    template_users = console.users.findall(type='template')
    if not template_users:
        skip_warn(f"No template users on HMC {hd.host}")
    template_user = template_users[0]

    # Test creating the user pattern

    upatt_input_props = {
        'name': upatt_name,
        'description': 'Test user pattern for zhmcclient end2end tests',
        'pattern': TEST_PREFIX + ' test_upatt_crud .+',
        'type': 'regular-expression',
        'retention-time': 180,
        'specific-template-uri': template_user.uri,
    }
    upatt_auto_props = {}

    # The code to be tested
    try:
        upatt = console.user_patterns.create(upatt_input_props)
    except zhmcclient.HTTPError as exc:
        if exc.http_status == 403 and exc.reason == 1:
            skip_warn(
                f"HMC userid {hd.userid!r} is not authorized for task "
                f"'Manage User Patterns' on HMC {hd.host}")
        else:
            raise

    for pn, exp_value in upatt_input_props.items():
        assert upatt.properties[pn] == exp_value, \
            f"Unexpected value for property {pn!r}"
    upatt.pull_full_properties()
    for pn, exp_value in upatt_input_props.items():
        assert upatt.properties[pn] == exp_value, \
            f"Unexpected value for property {pn!r}"
    for pn, exp_value in upatt_auto_props.items():
        assert upatt.properties[pn] == exp_value, \
            f"Unexpected value for property {pn!r}"

    # Test updating a property of the user pattern

    new_desc = "Updated user pattern description."

    # The code to be tested
    upatt.update_properties(dict(description=new_desc))

    assert upatt.properties['description'] == new_desc
    upatt.pull_full_properties()
    assert upatt.properties['description'] == new_desc

    # Test renaming the user pattern

    # The code to be tested
    upatt.update_properties(dict(name=upatt_name_new))

    assert upatt.properties['name'] == upatt_name_new
    upatt.pull_full_properties()
    assert upatt.properties['name'] == upatt_name_new
    with pytest.raises(zhmcclient.NotFound):
        console.user_patterns.find(name=upatt_name)

    # Test deleting the user pattern

    # The code to be tested
    upatt.delete()

    with pytest.raises(zhmcclient.NotFound):
        console.user_patterns.find(name=upatt_name_new)



# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for LPARs (on CPCs in DPM mode).

These tests do not change any existing LPARs, but create, modify and delete
test LPARs.
"""


import random
from datetime import timedelta, datetime, timezone
import pdb
import pytest
from requests.packages import urllib3

import zhmcclient
# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401, E501
from zhmcclient.testutils import classic_mode_cpcs  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import

from .utils import skip_warn, pick_test_resources, runtest_find_list, \
    runtest_get_properties, ensure_lpar_inactive, set_resource_property, \
    setup_logging

urllib3.disable_warnings()

# Print debug messages in tests
DEBUG = False

# Logging for zhmcclient HMC interactions and test functions
LOGGING = False
LOG_FILE = 'test_lpar.log'

# Properties in minimalistic Lpar objects (e.g. find_by_name())
LPAR_MINIMAL_PROPS = ['object-uri', 'name']

# Properties in Lpar objects returned by list() without full props
LPAR_LIST_PROPS = ['object-uri', 'name', 'status']

# Properties returned by default from list_permitted_lpars()
LPAR_LIST_PERMITTED_PROPS = [
    'name', 'object-uri', 'activation-mode', 'status',
    'has-unacceptable-status', 'cpc-name', 'cpc-object-uri',
    # HMCs return 'se-version' on HMC API version 4.10 or higher
]


# Properties whose values can change between retrievals of Lpar objects
LPAR_VOLATILE_PROPS = []


def test_lpar_find_list(classic_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test list(), find(), findall().
    """
    if not classic_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in classic mode")

    for cpc in classic_mode_cpcs:
        assert not cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the LPARs to test with
        lpar_list = cpc.lpars.list()
        if not lpar_list:
            skip_warn(f"No LPARs on CPC {cpc.name} managed by HMC {hd.host}")
        lpar_list = pick_test_resources(lpar_list)

        for lpar in lpar_list:
            print(f"Testing on CPC {cpc.name} with LPAR {lpar.name!r}")
            runtest_find_list(
                session, cpc.lpars, lpar.name, 'name', 'status',
                LPAR_VOLATILE_PROPS, LPAR_MINIMAL_PROPS, LPAR_LIST_PROPS)


def test_lpar_property(classic_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test property related methods
    """
    if not classic_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in classic mode")

    for cpc in classic_mode_cpcs:
        assert not cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the LPARs to test with
        lpar_list = cpc.lpars.list()
        if not lpar_list:
            skip_warn(f"No LPARs on CPC {cpc.name} managed by HMC {hd.host}")
        lpar_list = pick_test_resources(lpar_list)

        for lpar in lpar_list:
            print(f"Testing on CPC {cpc.name} with LPAR {lpar.name!r}")

            # Select a property that is not returned by list()
            non_list_prop = 'description'

            runtest_get_properties(lpar.manager, non_list_prop)


TESTCASES_CONSOLE_LIST_PERMITTED_LPARS = [
    # Testcases for test_console_list_permitted_lpars()

    # Each testcase is a tuple of:
    # * desc: description
    # * input_kwargs: Keyword arguments to the test function
    # * exp_prop_names: List of expected property names
    (
        "Default arguments",
        {},
        LPAR_LIST_PERMITTED_PROPS,
    ),
    (
        "Explicit defaults for arguments",
        {
            'full_properties': False,
            'filter_args': None,
            'additional_properties': None
        },
        LPAR_LIST_PERMITTED_PROPS,
    ),
    (
        "One server-side filter",
        {
            'filter_args': {'status': 'not-operating'},
        },
        LPAR_LIST_PERMITTED_PROPS,
    ),
    (
        "One client-side filter that is not in result",
        {
            'filter_args': {'has-operating-system-messages': False},
        },
        LPAR_LIST_PERMITTED_PROPS,
    ),
    (
        "Empty additional properties",
        {
            'additional_properties': [],
        },
        LPAR_LIST_PERMITTED_PROPS,
    ),
    (
        "One additional property",
        {
            'additional_properties': ['description'],
        },
        LPAR_LIST_PERMITTED_PROPS + ['description'],
    ),
    (
        "Full properties",
        {
            'full_properties': True,
        },
        LPAR_LIST_PERMITTED_PROPS + ['has-operating-system-messages'],
    ),
]


@pytest.mark.parametrize(
    "desc, input_kwargs, exp_prop_names",
    TESTCASES_CONSOLE_LIST_PERMITTED_LPARS)
def test_console_list_permitted_lpars(
        classic_mode_cpcs, desc, input_kwargs, exp_prop_names):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test Console.list_permitted_lpars() method
    """
    if not classic_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in classic mode")

    logger = setup_logging(LOGGING, 'test_console_list_permitted_lpars',
                           LOG_FILE)
    logger.debug("Entered test function with: "
                 "desc=%r, input_kwargs=%r, exp_prop_names=%r",
                 desc, input_kwargs, exp_prop_names)

    for cpc in classic_mode_cpcs:
        assert not cpc.dpm_enabled

        logger.debug("Testing with CPC %s", cpc.name)

        client = cpc.manager.client
        console = client.consoles.console

        logger.debug("Calling list_permitted_lpars() with kwargs: %r",
                     input_kwargs)

        # Execute the code to be tested
        lpars = console.list_permitted_lpars(**input_kwargs)

        logger.debug("list_permitted_lpars() returned %d LPARs", len(lpars))

        for lpar in lpars:
            lpar_props = dict(lpar.properties)
            for pname in exp_prop_names:
                assert pname in lpar_props, (
                    f"Property {pname!r} missing from returned LPAR "
                    f"properties, got: {lpar_props!r}")

    logger.debug("Leaving test function")


def test_lpar_list_os_messages(classic_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test "List OS Messages" operation on LPARs
    """
    if not classic_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in classic mode")

    for cpc in classic_mode_cpcs:
        assert not cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        if hd.mock_file:
            skip_warn("zhmcclient mock does not support 'List OS Messages' "
                      "operation")

        # Pick the LPAR to test with
        lpar_list = cpc.lpars.list()
        if not lpar_list:
            skip_warn(f"No LPARs on CPC {cpc.name} managed by HMC {hd.host}")

        test_lpar = None
        for lpar in lpar_list:

            # Test: List all messages (without begin or end)
            try:
                if DEBUG:
                    print("Debug: Test: Listing OS messages of LPAR "
                          f"{lpar.name} with no begin/end")
                result = lpar.list_os_messages()
            except zhmcclient.HTTPError as exc:
                if exc.http_status == 409 and exc.reason == 332:
                    # Meaning: The messages interface for the LPAR is not
                    # available
                    if DEBUG:
                        print(f"Debug: LPAR {lpar.name} cannot list OS "
                              "messages")
                    continue
                raise

            all_messages = result['os-messages']
            if len(all_messages) >= 3:
                test_lpar = lpar
                break

            if DEBUG:
                print(f"Debug: LPAR {lpar.name} has only {len(all_messages)} "
                      "OS messages")

        if test_lpar is None:
            skip_warn(f"No LPAR on CPC {cpc.name} has the minimum number of "
                      "3 OS messages for the test")

        # Test with begin/end selecting the full set of messages
        all_begin = all_messages[0]['sequence-number']
        all_end = all_messages[-1]['sequence-number']
        if DEBUG:
            print(f"Debug: Test: Listing OS messages of LPAR {test_lpar.name} "
                  f"with begin={all_begin}, end={all_end}")
        result = test_lpar.list_os_messages(begin=all_begin, end=all_end)
        messages = result['os-messages']
        assert len(messages) == all_end - all_begin + 1
        assert messages == all_messages

        # Test with begin/end selecting a subset of messages
        while True:
            seq1 = random.choice(all_messages)['sequence-number']
            if seq1 not in {all_begin, all_end}:
                break
        while True:
            seq2 = random.choice(all_messages)['sequence-number']
            if seq2 not in {all_begin, all_end, seq1}:
                break
        begin = min(seq1, seq2)
        end = max(seq1, seq2)
        if DEBUG:
            print(f"Debug: Test: Listing OS messages of LPAR {test_lpar.name} "
                  f"with begin={begin}, end={end}")
        result = test_lpar.list_os_messages(begin=begin, end=end)
        messages = result['os-messages']
        assert len(messages) == end - begin + 1
        for message in messages:
            assert message in all_messages

        # Test with begin/end and maximum messages
        max_messages = random.randint(0, end - begin + 1)
        if DEBUG:
            print(f"Debug: Test: Listing OS messages of LPAR {test_lpar.name} "
                  f"with begin={begin}, end={end}, max_messages={max_messages}")
        result = test_lpar.list_os_messages(
            begin=begin, end=end, max_messages=max_messages)
        messages = result['os-messages']
        assert len(messages) == max_messages
        for message in messages:
            assert message in all_messages

        # Test with is_held
        for is_held in (False, True):
            if DEBUG:
                print("Debug: Test: Listing OS messages of LPAR "
                      f"{test_lpar.name} with is_held={is_held}")
            result = test_lpar.list_os_messages(is_held=is_held)
            messages = result['os-messages']
            for message in messages:
                assert message['is-held'] == is_held

        # Test with is_priority
        for is_priority in (False, True):
            if DEBUG:
                print("Debug: Test: Listing OS messages of LPAR "
                      f"{test_lpar.name} with is_priority={is_priority}")
            result = test_lpar.list_os_messages(is_priority=is_priority)
            messages = result['os-messages']
            for message in messages:
                assert message['is-priority'] == is_priority


LPAR_ACTIVATE_TESTCASES = [
    # Testcases for test_lpar_activate().
    # The list items are tuples with the following items:
    # - desc (string): description of the testcase.
    # - lpar_mode (string): Activation mode of the LPAR.
    # - ap_type (string): Type of activation profile to be used for
    #   activation_profile_name parm in Lpar.activate(): 'image', 'wrongimage',
    #   'load', None.
    # - nap_type (string): Type of activation profile to be used for
    #   'next-activation-profile-name' prop of LPAR: 'image', 'load'.
    # - auto_load (bool): Value for 'load-at-activation' in image profile.
    # - input_kwargs (dict): Input parameters for Lpar.activate() in addition
    #   to activation_profile_name.
    # - exp_props (dict): Dict of expected properties of the Lpar after
    #   calling Lpar.activate().
    # - exp_exc_type (class): Expected exception type, or None for success.
    # - run (bool or 'pdb'): Whether to run the test or call the debugger.

    (
        "General LPAR w/o auto-load, default parms, next profile is image",
        'general',
        None,
        'image',
        False,
        dict(),
        {
            'status': 'not-operating',
        },
        None,
        True,
    ),
    (
        "General LPAR with auto-load, default parms, next profile is image",
        'general',
        None,
        'image',
        True,
        dict(),
        {
            'status': 'operating',
        },
        None,
        True,
    ),
    (
        "General LPAR w/o auto-load, image profile as parm",
        'general',
        'image',
        'image',
        False,
        dict(),
        {
            'status': 'not-operating',
        },
        None,
        True,
    ),
    (
        "Linux LPAR with auto-load, image profile as parm",
        'linux',
        'image',
        'image',
        True,
        dict(),
        {
            'status': 'operating',
        },
        None,
        True,
    ),

    (
        "Linux LPAR w/o auto-load, default parms, next profile is image",
        'linux',
        None,
        'image',
        False,
        dict(),
        {
            'status': 'not-operating',
        },
        None,
        True,
    ),
    (
        "Linux LPAR with auto-load, default parms, next profile is image",
        'linux',
        None,
        'image',
        True,
        dict(),
        {
            'status': 'operating',
        },
        None,
        True,
    ),
    (
        "Linux LPAR w/o auto-load, image profile as parm",
        'linux',
        'image',
        'image',
        False,
        dict(),
        {
            'status': 'not-operating',
        },
        None,
        True,
    ),
    (
        "Linux LPAR with auto-load, image profile as parm",
        'linux',
        'image',
        'image',
        True,
        dict(),
        {
            'status': 'operating',
        },
        None,
        True,
    ),

    (
        "SSC LPAR (always auto-loads), default parms, next profile is image",
        'ssc',
        None,
        'image',
        False,
        dict(),
        {
            'status': 'operating',
        },
        None,
        True,
    ),
    (
        "SSC LPAR (always auto-loads), image profile as parm",
        'ssc',
        'image',
        'image',
        False,
        dict(),
        {
            'status': 'operating',
        },
        None,
        True,
    ),
    (
        "Linux LPAR, wrong image profile",
        'linux',
        'wrongimage',
        'image',
        False,
        dict(),
        None,
        # HTTP 500,263: LPAR name and image profile name must match
        zhmcclient.HTTPError,
        True,
    ),
    (
        "SSC LPAR, wrong image profile",
        'ssc',
        'wrongimage',
        'image',
        False,
        dict(),
        None,
        # HTTP 500,263: LPAR name and image profile name must match
        zhmcclient.HTTPError,
        True,
    ),

    # TODO: Add testcases with load profiles
]


@pytest.mark.parametrize(
    "desc, lpar_mode, ap_type, nap_type, auto_load, input_kwargs, "
    "exp_props, exp_exc_type, run",
    LPAR_ACTIVATE_TESTCASES)
def test_lpar_activate(
        desc, lpar_mode, ap_type, nap_type, auto_load, input_kwargs, exp_props,
        exp_exc_type, run, classic_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name, unused-argument
    """
    Test Lpar.activate().

    Note: These tests require that the following properties are present in the
    HMC inventory file for the CPC item of the HMC that is tested against:

    * 'loadable_lpars' - The names of LPARs for some operating modes, that
      properly reach the 'operating' status when activated under auto-load
      conditions. If this item is missing, all tests will be skipped.
      If the entries for certain operating modes are missing, the tests
      for these operating modes are skipped.

    * 'load_profiles' - The names of load profiles that can be used to properly
      load the corresponding partitions in 'loadable_lpars'. If this item is
      missing, all tests will be skipped. If the entries for certain operating
      modes are missing, the load-profile related tests for these operating
      modes are skipped.

    Example::

        A01:
          . . .
          cpcs:
            P0000A01:           # <- CPC item of the HMC that is tested against
              machine_type: "3931"
              machine_model: "A01"
              dpm_enabled: false
              loadable_lpars:   # <- specific entry for this test
                general: "LP01"      # key: operating mode, value: LPAR name
                linux-only: "LP02
                ssc: "LP03"
              load_profiles:    # <- specific entry for this test
                general: "LP01LOAD"  # key: op. mode, value: load profile name
                linux-only: "LP02LOAD"
                ssc: "LP03LOAD"

    """
    if not classic_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in classic mode")

    if not run:
        skip_warn("Testcase is disabled in testcase definition")

    logger = setup_logging(LOGGING, 'test_lpar_activate', LOG_FILE)

    for cpc in classic_mode_cpcs:
        assert not cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        # The following has no check since classic_mode_cpcs only contains
        # CPCs that have such an item:
        hd_cpc = hd.cpcs[cpc.name]

        try:
            loadable_lpars = hd_cpc['loadable_lpars']
        except KeyError:
            pytest.skip(
                f"Inventory file entry for HMC nickname {hd.nickname!r} does "
                "not have a 'loadable_lpars' property in its entry for CPC "
                f"{cpc.name!r}")
        try:
            load_profiles = hd_cpc['load_profiles']
        except KeyError:
            pytest.skip(
                f"Inventory file entry for HMC nickname {hd.nickname!r} does "
                "not have a 'load_profiles' property in its entry for "
                f"CPC {cpc.name!r}")

        try:
            lpar_name = loadable_lpars[lpar_mode]
        except (KeyError, TypeError):
            pytest.skip(
                f"Inventory file entry for HMC nickname {hd.nickname!r} does "
                f"not have an entry for operating mode {lpar_mode!r} in its "
                f"'loadable_lpars' property for CPC {cpc.name!r}")

        # Find the image profile corresponding to the LPAR, and the other
        # (wrong) image profile names for specific tests with that.
        iap_name = lpar_name
        all_iaps = cpc.image_activation_profiles.list()
        lpar_iaps = [_iap for _iap in all_iaps if _iap.name == lpar_name]
        if len(lpar_iaps) >= 1:
            iap = lpar_iaps[0]
        else:
            pytest.skip(f"Image activation profile {iap_name!r} does not exist "
                        f"on CPC {cpc.name}.")
        wrong_iap_names = [_iap.name for _iap in all_iaps
                           if _iap.name != lpar_name]

        if ap_type == 'image':
            ap_name = lpar_name
        if ap_type == 'wrongimage':
            ap_name = random.choice(wrong_iap_names)
        elif ap_type == 'load':
            try:
                ap_name = load_profiles[lpar_mode]
            except (KeyError, TypeError):
                pytest.skip(
                    f"Inventory file entry for HMC nickname {hd.nickname!r} "
                    f"does not have an entry for operating mode {lpar_mode!r} "
                    f"in its 'load_profiles' property for CPC {cpc.name!r}")
        else:
            ap_name = None

        if nap_type == 'image':
            nap_name = lpar_name
        elif nap_type == 'load':
            try:
                nap_name = load_profiles[lpar_mode]
            except (KeyError, TypeError):
                pytest.skip(
                    f"Inventory file entry for HMC nickname {hd.nickname!r} "
                    f"does not have an entry for operating mode {lpar_mode!r} "
                    f"in its 'load_profiles' property for CPC {cpc.name!r}")
        else:
            nap_name = None

        try:
            lpar = cpc.lpars.find(name=lpar_name)
        except zhmcclient.NotFound:
            pytest.skip(f"LPAR {lpar_name!r} does not exist on CPC {cpc.name}.")

        if ap_type == 'load' or nap_type == 'load':
            lap_name = load_profiles[lpar_mode]
            try:
                cpc.load_activation_profiles.find(name=lap_name)
            except zhmcclient.NotFound:
                pytest.skip(f"Load activation profile {lap_name!r} does not "
                            f"exist on CPC {cpc.name}.")

        # pylint: disable=possibly-used-before-assignment
        op_mode = iap.get_property('operating-mode')
        assert op_mode == lpar_mode, (
            f"Incorrect testcase definition: Operating mode {op_mode!r} in "
            f"image activation profile {iap_name!r} on CPC {cpc.name} does not "
            f"match the lpar_mode {lpar_mode!r} of the testcase")

        msg = f"Testing on CPC {cpc.name} with LPAR {lpar.name!r}"
        print(msg)
        logger.info(msg)

        if run == 'pdb':
            # pylint: disable=forgotten-debug-statement
            pdb.set_trace()

        logger.info("Preparation: Ensuring that LPAR %r is inactive",
                    lpar.name)
        ensure_lpar_inactive(lpar)

        logger.info("Preparation: Setting 'next-activation-profile-name' = %r "
                    "in LPAR %r", nap_name, lpar.name)
        saved_nap_name = set_resource_property(
            lpar, 'next-activation-profile-name', nap_name)

        logger.info("Preparation: Setting 'load-at-activation' = %r in image "
                    "profile %r", auto_load, iap.name)
        saved_auto_load = set_resource_property(
            iap, 'load-at-activation', auto_load)

        try:
            logger.info("Test: Activating LPAR %r (profile arg: %r, "
                        "add. args: %r)", lpar.name, ap_name, input_kwargs)

            if exp_exc_type:

                with pytest.raises(exp_exc_type):

                    # Exercise the code to be tested
                    lpar.activate(
                        activation_profile_name=ap_name, **input_kwargs)

            else:

                # Exercise the code to be tested
                lpar.activate(
                    activation_profile_name=ap_name, **input_kwargs)

                # In case of an image profile with auto-load, the activate()
                # method waits until status 'operating' is reached. If that
                # cannot be reached, StatusTimeout is raised.

                # Check the expected properties
                lpar.pull_full_properties()
                lpar_props = dict(lpar.properties)
                logger.info("Status of LPAR %r is %r after test",
                            lpar.name, lpar_props['status'])
                for pname, exp_value in exp_props.items():
                    assert pname in lpar_props, (
                        f"Expected property {pname!r} does not exist in "
                        f"actual properties of LPAR {lpar.name!r}")
                    act_value = lpar_props[pname]
                    assert act_value == exp_value, (
                        f"Property {pname!r} has unexpected value "
                        f"{act_value!r}; expected value: {exp_value!r}")
        finally:
            logger.info("Cleanup: Ensuring that LPAR %r is inactive",
                        lpar.name)
            ensure_lpar_inactive(lpar)

            logger.info("Cleanup: Setting 'next-activation-profile-name' = %r "
                        "in LPAR %r", saved_nap_name, lpar.name)
            set_resource_property(
                lpar, 'next-activation-profile-name', saved_nap_name)

            logger.info("Cleanup: Setting 'load-at-activation' = %r in image "
                        "profile %r", saved_auto_load, iap.name)
            set_resource_property(iap, 'load-at-activation', saved_auto_load)


TESTCASES_LPAR_GET_SUSTAINABILITY_DATA = [
    # Test cases for test_lpar_get_sustainability_data(), each as a tuple with
    # these items:
    # * tc: Testcase short description
    # * input_kwargs: kwargs to be used as input parameters for
    #   Lpar.get_sustainability_data()
    # * exp_oldest: expected delta time from now to oldest data point,
    #   as timedelta
    # * exp_delta: expected delta time between data points, as timedelta
    (
        "Default values (range=last-week, resolution=one-hour)",
        {},
        timedelta(days=7),
        timedelta(hours=1),
    ),
    (
        "range=last-day, resolution=one-hour",
        {
            "range": "last-day",
            "resolution": "one-hour",
        },
        timedelta(hours=24),
        timedelta(hours=1),
    ),
    (
        "range=last-day, resolution=fifteen-minutes",
        {
            "range": "last-day",
            "resolution": "fifteen-minutes",
        },
        timedelta(hours=24),
        timedelta(minutes=15),
    ),
]

LPAR_METRICS = {
    # Metrics returned in "Get LPAR Historical Sustainability Data" HMC
    # operation.
    # metric name: data type
    "wattage": int,
    "processor-utilization": int,
}


@pytest.mark.parametrize(
    "tc, input_kwargs, exp_oldest, exp_delta",
    TESTCASES_LPAR_GET_SUSTAINABILITY_DATA)
def test_lpar_get_sustainability_data(
        tc, input_kwargs, exp_oldest, exp_delta,
        classic_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name,unused-argument
    """
    Test for Lpar.get_sustainability_data(...)
    """
    if not classic_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in classic mode")

    for cpc in classic_mode_cpcs:
        assert not cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        if hd.mock_file:
            skip_warn("zhmcclient mock does not support "
                      "Lpar.get_sustainability_data()")

        # Pick the LPAR to test with
        lpar_list = cpc.lpars.list()
        if not lpar_list:
            skip_warn(f"No LPARs on CPC {cpc.name} managed by HMC {hd.host}")

        # Pick a random LPAR to test with
        lpar = random.choice(lpar_list)

        session = lpar.manager.session
        hd = session.hmc_definition

        print(f"Testing with LPAR {lpar.name}")

        try:

            # The code to be tested
            data = lpar.get_sustainability_data(**input_kwargs)

        except zhmcclient.HTTPError as exc:
            if exc.http_status == 403 and exc.reason == 1:
                skip_warn(
                    f"HMC userid {hd.userid!r} is not authorized for task "
                    f"'Environmental Dashboard' on HMC {hd.host}")
            elif exc.http_status == 404 and exc.reason == 1:
                skip_warn(
                    f"LPAR {lpar.name} on HMC {hd.host} does not support "
                    f"feature: {exc}")
            else:
                raise

        now_dt = datetime.now(timezone.utc)
        exp_oldest_dt = now_dt - exp_oldest

        act_metric_names = set(data.keys())
        exp_metric_names = set(LPAR_METRICS.keys())
        assert act_metric_names == exp_metric_names

        for metric_name, metric_array in data.items():
            metric_type = LPAR_METRICS[metric_name]

            first_item = True
            previous_dt = None
            for dp_item in metric_array:
                # We assume the order is oldest to newest

                assert 'data' in dp_item
                assert 'timestamp' in dp_item
                assert len(dp_item) == 2

                dp_data = dp_item['data']
                dp_timestamp_dt = dp_item['timestamp']

                assert isinstance(dp_data, metric_type), \
                    f"Invalid data type for metric {metric_name!r}"

                if first_item:
                    first_item = False

                    # Verify that the oldest timestamp is within a certain
                    # delta from the range start.
                    # There are cases where that is not satisfied, so we only
                    # issue only a warning (as opposed to failing).
                    delta_sec = abs((dp_timestamp_dt - exp_oldest_dt).seconds)
                    if delta_sec > 15 * 60:
                        print(f"Warning: Oldest data point of metric "
                              f"{metric_name!r} is not within 15 minutes of "
                              "range start: Oldest data point: "
                              f"{dp_timestamp_dt}, Range start: "
                              f"{exp_oldest_dt}, Delta: {delta_sec} sec")
                else:

                    # For second oldest timestamp on, verify that the delta
                    # to the previous data point is the requested resolution.
                    # There are cases where that is not satisfied, so we only
                    # issue only a warning (as opposed to failing).
                    tolerance_pct = 10
                    delta_td = abs(dp_timestamp_dt - previous_dt)
                    if abs(delta_td.seconds - exp_delta.seconds) > \
                            tolerance_pct / 100 * exp_delta.seconds:
                        print("Warning: Timestamp of a data point of metric "
                              f"{metric_name!r} is not within expected delta "
                              "of its previous data point. Actual delta: "
                              f"{delta_td}, Expected delta: {exp_delta} "
                              f"(+/-{tolerance_pct}%)")

                previous_dt = dp_timestamp_dt
